import re
from typing import Optional

from ..lib.log_tools import make_default_logger
from ..lib.string_utils import clean_string, convert_date_string, force_unicode
from .reports import BaseReport
from .utils import is_pdf, make_doc1_url, make_docs1_url

logger = make_default_logger()


class DownloadConfirmationPage(BaseReport):
    """An object for querying and parsing appellate PACER documents confirmation
    download page.
    """

    def __init__(self, court_id, pacer_session=None):
        super().__init__(court_id, pacer_session)

        self.is_appellate = False
        if self.court_id[-1].isdigit() or self.court_id in [
            "cadc",
            "cafc",
            "cavc",
        ]:
            self.is_appellate = True

    def query(self, pacer_doc_id):
        """Query the "confirmation download page" endpoint and set the results
        to self.response.

        :param pacer_doc_id: The internal PACER document ID for the item.
        :return: a request response object
        """

        assert (
            self.session is not None
        ), "session attribute of DownloadConfirmationPage cannot be None."

        if self.is_appellate:
            # Make the appellate document URL
            url = make_docs1_url(self.court_id, pacer_doc_id, True)
        else:
            # Make the district/bankruptcy document URL
            url = make_doc1_url(self.court_id, pacer_doc_id, True)

        logger.info("Querying the confirmation page endpoint at URL: %s", url)
        self.response = self.session.get(url)
        if is_pdf(self.response):
            # Sometimes the PDF document is returned without showing the
            # download confirmation page, not a valid page to parse.
            self.is_valid = False
            return
        self.parse()

    @property
    def data(self):
        """Get data back from the query for the matching document entry.

        :return: If lookup fails, an empty dict. Else, a dict containing the
        following fields:
            - document_number: The document number we're working with
            - docket_number: The docket case number we're working with
            - cost: The document cost we're working with
            - billable_pages: The document billable pages we're working with
            - document_description: The document description we're working with
            - transaction_date: The document transaction date we're working with

        See the JSON objects in the tests for more examples.
        """
        if self.is_valid is False:
            return {}

        if not self._is_a_receipt_page():
            # Abort. If we cannot get a document number return a empy dict.
            # It's not a valid confirmation page.
            return {}

        return {
            "document_number": self._get_document_number(),
            "docket_number": self._get_docket_number(),
            "cost": self._get_document_cost(),
            "billable_pages": self._get_billable_pages(),
            "document_description": self._get_document_description(),
            "transaction_date": self._get_transaction_date(),
        }

    def _is_a_receipt_page(self) -> bool:
        """Check if this is a valid download confirmation page for a district
        bankruptcy or appellate court.

        :return: True if is a valid page, otherwise False.
        """

        try:
            transaction_str = self.tree.re_xpath(
                '//*[re:match(text(), "Transaction Receipt")]'
            )[0]
        except IndexError:
            return False
        return True

    def _get_document_number(self) -> Optional[str]:
        """Get the document number for an item.

        :return: The PACER document number if available, otherwise None.
        """

        try:
            document_and_case_number = self.tree.xpath(
                '//strong[contains(., "Document: PDF Document")]'
            )[0].text_content()
        except IndexError:
            return None

        regex = r", Document:([^\)]*)"
        document_number = re.findall(regex, document_and_case_number)
        if document_number:
            return clean_string(document_number[0])
        return None

    def _get_document_cost(self) -> Optional[str]:
        """Get the document cost for an item.

        :return: The PACER document cost if available, otherwise None.
        """
        try:
            cost_str = self.tree.re_xpath(
                '//*[re:match(text(), "Cost:")]/'
                "/ancestor::th[1]/following-sibling::td[1]/font[1]"
            )[0].text_content()
        except IndexError:
            return None

        if cost_str:
            return clean_string(cost_str)
        return None

    def _get_docket_number(self) -> Optional[str]:
        """Get the docket number for an item.

        :return: The PACER docket number if available, otherwise None.
        """

        try:
            if self.is_appellate:
                document_and_case_number = self.tree.xpath(
                    '//strong[contains(., "Document: PDF Document")]'
                )[0].text_content()
            else:
                docket_number = self.tree.re_xpath(
                    '//*[re:match(text(), "Case Number:")]/'
                    "/ancestor::th[1]/following-sibling::td[1]/font[1]"
                )[0].text_content()

        except IndexError:
            return None

        if self.is_appellate:
            regex = r"Case:([^\,]*)"
            docket_number = re.findall(regex, document_and_case_number)[0]

        if docket_number:
            return clean_string(docket_number)
        return None

    def _get_billable_pages(self) -> Optional[str]:
        """Get the document billable pages.

        :return: The document billable pages if available, otherwise None.
        """
        try:
            billable_pages_str = self.tree.re_xpath(
                '//*[re:match(text(), "Billable Pages:")]/'
                "/ancestor::th[1]/following-sibling::td[1]/font[1]"
            )[0].text_content()
        except IndexError:
            return None

        if billable_pages_str:
            return clean_string(billable_pages_str)
        return None

    def _get_document_description(self) -> Optional[str]:
        """Get the document description for an item.

        :return: The PACER document description if available, otherwise None.
        """
        try:
            document_description_str = self.tree.re_xpath(
                '//*[re:match(text(), "Description:")]/'
                "/ancestor::th[1]/following-sibling::td[1]/font[1]"
            )[0].text_content()
        except IndexError:
            return None

        if document_description_str:
            return clean_string(document_description_str)
        return None

    def _get_transaction_date(self) -> Optional[str]:
        """Get the PACER transaction date.

        :return: The PACER transaction date if available, otherwise None.
        """

        try:
            transaction_date_str = self.tree.re_xpath(
                '//*[re:match(text(), "Transaction Receipt")]/'
                "/ancestor::tr[1]/following-sibling::tr[3]"
            )[0].text_content()
        except IndexError:
            return None

        if "-" in transaction_date_str:
            # Some courts include additional data besides the date time:
            # 5th Circuit - Appellate - 08/30/2022 13:48:21
            # 2th Circuit - 08/30/2022 13:48:21
            # datetime it's always at the end.
            transaction_date_str = transaction_date_str.split("-")[-1]

        transaction_date_str = force_unicode(transaction_date_str)
        transaction_date_str = convert_date_string(
            transaction_date_str, datetime=True
        )

        if transaction_date_str:
            return transaction_date_str
        return None



from datetime import date, datetime, timedelta

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "http://www.courts.state.va.us/scndex.htm"
        self.status = "Published"

    def _process_html(self):
        if self.test_mode_enabled():
            today = datetime.strptime("11/20/2023", "%m/%d/%Y").date()
        else:
            today = date.today()

        for row in self.html.xpath("//p"):
            links = row.xpath(".//a")
            if len(links) != 2:
                continue
            name = row.xpath(".//b/text()")[0].strip()
            summary = row.text_content().split(name)[1].strip()
            date_str = summary.split("\n")[0].strip()
            if "Revised" in date_str:
                date_str = date_str.split("Revised")[1].strip().strip(")")
            date_object = datetime.strptime(date_str, "%m/%d/%Y").date()
            if date_object < today - timedelta(days=60):
                # Stop after two months
                break

            self.cases.append(
                {
                    "name": row.xpath(".//b/text()")[0].strip(),
                    "docket": links[0].get("name"),
                    "url": links[1].get("href"),
                    "summary": summary,
                    "date": date_str,
                }
            )



"""
Scraper for Alabama Supreme Court
CourtID: ala
Court Short Name: Alabama
Author: William Palin
Court Contact:
History:
 - 2023-01-04: Created.
 - 2023-011-14: Alabama no longer uses page or use selenium.
"""

import json

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.court_str = "68f021c4-6a44-4735-9a76-5360b2e8af13"

    def _build_url(self, court_str):
        return f"https://publicportal-api.alappeals.gov/courts/cms/publications?courtID={court_str}&page=0&size=25&sort=publicationDate%2Cdesc"

    def _download(self, request_dict={}):
        if self.test_mode_enabled():
            with open(self.url) as file:
                self.json = json.load(file)
        else:
            self.url = self._build_url(self.court_str)
            self.json = super()._download()

    def _process_html(self):
        for item in self.json["_embedded"]["results"][:1]:
            date_filed = item["scheduledDate"][:10]
            for publicationItem in item["publicationItems"]:
                if not publicationItem.get("documents", []):
                    continue

                url = f"https://publicportal-api.alappeals.gov/courts/{self.court_str}/cms/case/{publicationItem['caseInstanceUUID']}/docketentrydocuments/{publicationItem['documents'][0]['documentLinkUUID']}"
                docket = publicationItem["caseNumber"]
                author = publicationItem["groupName"]
                name = publicationItem["title"]
                self.cases.append(
                    {
                        "date": date_filed,
                        "name": name,
                        "docket": docket,
                        "status": "Published",
                        "url": url,
                        "judge": author,
                    }
                )



"""Backscraper for Dept of Justice Office of Legal Counsel
CourtID: olc
Court Short Name: OLC
Author: William E. Palin
History:
    2022-01-14: Created by William E. Palin
"""

from juriscraper.opinions.united_states.administrative_agency import olc


class Site(olc.Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.back_scrape_iterable = range(1, 36)

    def _download_backwards(self, page: int) -> None:
        """Download backwards over the volume links in the DOJ OLC website.

        :param page: The page index for the URL
        :return: None
        """
        self.url = f"https://www.justice.gov/olc/opinions?items_per_page=40&page={page}"
        self.html = self._download()
        self._process_html()



from juriscraper.opinions.united_states.state import mo


class Site(mo.Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.court = "Southern"
        self.url = self.build_url()

    def _download(self, request_dict={}):
        # I have to do something to open up the page - but only for the southern
        # district of this verison of the court.
        if not self.test_mode_enabled():
            data = '{"sort_date_facets_by_value":true,"content_sample_length":300,"name":"default","path":"search","paging_states":[],"query_context":{"query_id":"1699132997921","prev_query_id":"1699132997873"},"user_context":{"referer":"https://www.courts.mo.gov/page.jsp?id=12086&dist=Opinions%20Southern&date=all&year=2023#all","locale":"en","service_id":"https://www.courts.mo.gov/search/apps/scripts/../../api/v2/","utc_time_zone_differential_in_seconds":-14400,"service_url":"https://www.courts.mo.gov/search/apps/scripts/../../api/v2/search"}}'
            self.request["session"].post(
                "https://www.courts.mo.gov/search/api/v2/search", data=data
            )
            self.method = "GET"
            self.request["session"].get(self.url)
        return super()._download(request_dict)

    def _process_html(self):
        for row in self.html.xpath("//div[@class='margin-bottom-15']"):
            date = row.xpath(".//input")[0].value
            for opinion in row.xpath(".//div[@class='list-group-item-text']"):
                url = opinion.xpath("a")[0].get("href")
                all_text = opinion.xpath(".//text()")
                case_metadata = [t.strip() for t in all_text if t.strip()]
                if len(case_metadata) != 6:
                    continue
                docket, name, _, author, _, vote = case_metadata
                self.cases.append(
                    {
                        "name": name,
                        "docket": docket[:-1],
                        "url": url,
                        "date": date,
                        "dispostion": vote.split(".")[0].strip(),
                        "judge": author,
                        "judges": vote.split(".", 1)[1].strip(),
                    }
                )



"""Scraper for the United States Court of Appeals for Veterans Claims
CourtID: cavc
Court Short Name: Vet. App.
History:
 - 2012-06-07: Created by Brian Carver
 - 2014-08-06: Updated by mlr.
 - 2023-01-23: Update by William Palin
"""

import datetime
import re
from datetime import date
from typing import Any, Dict

from juriscraper.lib.string_utils import titlecase
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = "http://www.uscourts.cavc.gov/opinions.php"
        self.court_id = self.__module__
        self.last_month = date.today() - datetime.timedelta(weeks=4)
        self.status = "Published"

    def _process_html(self):
        """Process the CAVC website and collect new opinions

        :return: None
        """
        if self.test_mode_enabled():
            self.last_month = datetime.datetime(2022, 12, 27).date()
        cases = self.html.xpath(".//tbody/tr/td/a/parent::td/parent::tr")
        for case in cases:
            try:
                _, date = case.xpath(".//td/text()")
                clean_date = datetime.datetime.strptime(date, "%d%b%y")
                if self.last_month > clean_date.date():
                    break
                self.cases.append(
                    {
                        "url": case.xpath(".//a/@href")[0],
                        "date": clean_date.strftime("%m/%d/%Y"),
                        "docket": case.xpath(".//a/text()")[0],
                        "name": "Case name extracted from text",
                    }
                )
            except ValueError:
                # The table has a malformed row
                continue

    def extract_from_text(self, scraped_text: str) -> Dict[str, Any]:
        """Can we extract the case name and clean it up?

        This method is a bit ... ugly.  Mostly due to bad PDFs that occur
        not often. This means weird whitespacing that makes this the most
        efficient method for name extraction

        :param scraped_text: The content of the document downloaded
        :return: Metadata to be added to the case
        """
        keepers = []
        start = False
        for row in scraped_text.split("\n"):
            check_row = re.sub(r"\s", "", row.upper())
            if re.findall(r"NOS?\.?\d+-\d+", check_row):
                start = True
                continue
            if not start:
                continue
            if "SECRETARY" in row.upper() or "VETERANS AFFAIRS" in row.upper():
                break
            if "Before" in row:
                break
            if not row.strip():
                continue
            if "JR" not in row:
                keepers.append(row.split(",", 1)[0].strip())
            else:
                keepers.append(row[: row.index("JR.") + 3].strip())
        case_name = titlecase(" ".join(keepers))
        metadata = {
            "OpinionCluster": {
                "case_name": case_name,
            },
        }
        return metadata



"""Scraper for Connecticut Appellate Court
CourtID: connctapp
Court Short Name: Connappct.
Author: Asadullah Baig<asadullahbeg@outlook.com>
Date created: 2014-07-11
History:
    - 2022-02-02, satsuki-chan: Updated to Opinionsitelinear
    - 2023-11-20, William Palin: Updated
"""

from datetime import date

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.year = date.today().strftime("%y")
        self.url = f"http://www.jud.ct.gov/external/supapp/archiveAROap{self.year}.htm"
        self.status = "Published"

    def _process_html(self) -> None:
        """Process the html and extract out the opinions

        :return: None
        """
        for date_section in self.html.xpath("//ul"):
            b = date_section[0].xpath(".//preceding::b[1]")[0]
            date = b.text_content().strip().split("of")[-1].strip()[:-1]
            for li in date_section.xpath(".//li"):
                link = li.xpath(".//a")[0]
                link_text = link.text_content()
                docket = link.text_content()
                name = li.text_content().replace(link_text, "").strip()
                self.cases.append(
                    {
                        "date": date,
                        "url": link.get("href"),
                        "docket": docket,
                        "name": name,
                    }
                )



"""Scraper for the New York Attorney General
CourtID: nyag
Court Short Name: New York Attorney General
"""

import datetime

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.year = datetime.date.today().year
        self.base_url = "https://ag.ny.gov/appeals-and-opinions/numerical-index?field_opinion_year_value=%d"
        self.url = self.base_url % self.year
        self.status = "Published"

    def _process_html(self):
        """"""
        for row in self.html.xpath("//div[@class='views-row']"):
            docket, _, _, summary, *_ = row.xpath(".//div/text()")
            url = row.xpath(".//div/span/p/a")[0].get("href")
            case = row.xpath(".//div/span/p/a/text()")[0]
            self.cases.append(
                {
                    "name": case,
                    "docket": docket,
                    "url": url,
                    "summary": summary,
                    "date": f"{docket[:4]}-07-01",
                    "date_filed_is_approximate": True,
                }
            )



import re
from itertools import chain, islice, tee

from .string_utils import force_unicode


def previous_and_next(some_iterable):
    """Provide previous and next values while iterating a list.

    This is from: http://stackoverflow.com/a/1012089/64911

    This will allow you to lazily iterate a list such that as you iterate, you
    get a tuple containing the previous, current, and next value.
    """
    prevs, items, nexts = tee(some_iterable, 3)
    prevs = chain([None], prevs)
    nexts = chain(islice(nexts, 1, None), [None])
    return zip(prevs, items, nexts)


def clean_court_object(obj):
    """Clean a list or dict that is part of a scraping response.

    Court data is notoriously horrible, so this function attempts to clean up
    common problems that it may have. You can pass in either a dict or a list,
    and it will be cleaned recursively.

    Supported cleanup includes:

    1. Removing spaces before commas.
    1. Stripping whitespace from the ends.
    1. Normalizing white space.
    1. Forcing unicode.

    :param obj: A dict or list containing string objects.
    :return: A dict or list with the string values cleaned.
    """
    if isinstance(obj, list):
        l = []
        for i in obj:
            l.append(clean_court_object(i))
        return l
    elif isinstance(obj, dict):
        d = {}
        for k, v in obj.items():
            d[k] = clean_court_object(v)
        return d
    elif isinstance(obj, str):
        s = " ".join(obj.strip().split())
        s = force_unicode(s)
        return re.sub(r"\s+,", ",", s)
    else:
        return obj



# Scraper for the United States Court of International Trade
# CourtID: cit
# Court Short Name: Ct. Int'l Trade
# Neutral Citation Format: Ct. Int'l Trade No. 12-1

import time
from datetime import date

from lxml import html

from juriscraper.OpinionSite import OpinionSite


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = "http://www.cit.uscourts.gov/SlipOpinions/index.html"
        self.court_id = self.__module__
        self.base = '//tr[../tr/th[contains(., "Caption")]]'

    def _get_download_urls(self):
        return [t for t in self.html.xpath(f"{self.base}/td[1]/a/@href")]

    def _get_citations(self):
        neutral_citations = []
        for t in self.html.xpath(f"{self.base}/td[1]/a/text()"):
            year, item_number = t.split("-")
            neutral_citations.append(f"20{year} CIT {item_number}")
        return neutral_citations

    def _get_case_names(self):
        # Exclude confidential rows by ensuring there is a sibling row that
        # contains an anchor (which confidential cases do not)
        case_names = []
        path = f"{self.base}/td[2][../td/a]"
        for e in self.html.xpath(path):
            text = e.text_content().strip()
            case_name = text.split("\n")[0]
            case_names.append(case_name)
        return case_names

    def _get_precedential_statuses(self):
        statuses = []
        for e in self.html.xpath(f"{self.base}/td[2][../td/a]"):
            s = (
                html.tostring(e, method="text", encoding="unicode")
                .lower()
                .strip()
            )
            if "errata" in s:
                statuses.append("Errata")
            else:
                statuses.append("Published")
        return statuses

    def _get_case_dates(self):
        # This does not capture the release dates for the errata documents.
        # The errata release date is listed in column 2. This will use the
        # original release date instead.
        dates = []
        date_formats = ["%m/%d/%Y", "%m/%d/%y"]
        for date_string in self.html.xpath(
            f"{self.base}/td[3][../td/a]//text()"
        ):
            for date_format in date_formats:
                try:
                    d = date.fromtimestamp(
                        time.mktime(
                            time.strptime(date_string.strip(), date_format)
                        )
                    )
                    dates.append(d)
                    break
                except ValueError:
                    # Try the next format
                    continue
        return dates

    def _get_docket_numbers(self):
        docket_numbers = []
        for e in self.html.xpath(f"{self.base}/td[4][../td/a]"):
            docket_numbers.append(
                html.tostring(e, method="text", encoding="unicode").strip()
            )
        return docket_numbers

    def _get_judges(self):
        judges = []
        for e in self.html.xpath(f"{self.base}/td[5][../td/a]"):
            s = html.tostring(e, method="text", encoding="unicode")
            judges.append(s)
        return judges

    def _get_nature_of_suit(self):
        return [
            t for t in self.html.xpath(f"{self.base}/td[6][../td/a]/text()")
        ]



"""Scraper for the Rhode Island Supreme Court
CourtID: ri
Court Short Name: R.I.
Court Contact: helpdesk@courts.ri.gov, MFerris@courts.ri.gov (Ferris, Mirella), webmaster@courts.ri.gov
    https://www.courts.ri.gov/PDF/TelephoneDirectory.pdf
Author: Brian W. Carver
Date created: 2013-08-10
History:
    Date created: 2013-08-10 by Brian W. Carver
    2022-05-02: Updated by William E. Palin, to use JSON responses
"""
from datetime import datetime

from lxml import html

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.status = "Published"
        self.opinion_type = "Opinions"
        self.url = self.build_url()

    def _download(self, request_dict={}):
        if self.test_mode_enabled():
            self.json = super()._download(request_dict)
            return
        return super()._download(request_dict)

    def fetch_json(self) -> None:
        """Fetch JSON data from the site.

        :return: None
        """
        if self.test_mode_enabled():
            return
        content = html.tostring(self.html)
        self.request["parameters"] = {
            "params": {
                "List": content.split(b"ctx.listName ")[1].split(b'"')[1],
                "View": content.split(b"ctx.view ")[1].split(b'"')[1],
                "ViewCount": 310,
                "IsXslView": "TRUE",
                "IsCSR": "TRUE",
            }
        }
        self._request_url_post(
            "https://www.courts.ri.gov/Courts/SupremeCourt/_layouts/15/inplview.aspx"
        )

    def _process_html(self) -> None:
        """Extract content from JSON response

        :return: None
        """
        self.fetch_json()
        for row in self.request["response"].json()["Row"]:
            first_docket = row["FileLeafRef"][:-4].split(",")[0]
            name = row["Title"].split(first_docket)[0].strip("Nos. ")
            self.cases.append(
                {
                    "date": row["Published_x0020_Date"],
                    "docket": row["FileLeafRef"][:-4],
                    "name": name.strip(","),
                    "summary": row.get("Description0", ""),
                    "url": f"https://www.courts.ri.gov{row['FileRef']}",
                }
            )

    def build_url(self) -> str:
        """Generate URL for R.I. Supreme Court

        # This court hears things from mid-September to end of June. This
        # defines the "term" for that year, which triggers the website updates.

        :return: URL to call before scraping JSON endpoint
        """
        today = datetime.today()
        this_year = today.year
        term_end = datetime(this_year, 9, 15)
        year = this_year if today >= term_end else this_year - 1
        return f"https://www.courts.ri.gov/Courts/SupremeCourt/Supreme{self.opinion_type}/Forms/{year}{year+1}.aspx"



"""
Scraper for Massachusetts Land Court
CourtID: massland
Court Short Name: Mass Land Ct
Author: William Palin
Court Contact:
Reviewer:
History:
 - 2020-12-05: Created.
Notes:
 - On Masscases.com they list all the cases on one long page.  A very atypical pattern
"""

import datetime
from typing import List
from urllib.parse import quote

from juriscraper.lib.string_utils import convert_date_string, titlecase
from juriscraper.OpinionSiteLinear import OpinionSite


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = "http://www.masscases.com/land_date.html"
        self.court_id = self.__module__
        self.year = None

    def _get_download_urls(self) -> List[str]:
        links = self.html.xpath(
            f'//*[@id="{self.year}"]/../following-sibling::section/table//tr/td[2]/a'
        )
        return [row.get("href") for row in links]

    def _get_docket_numbers(self) -> List[str]:
        links = self.html.xpath(
            f'//*[@id="{self.year}"]/../following-sibling::section/table//tr/td[2]/a'
        )
        return [row.text for row in links]

    def _get_case_names(self) -> List[str]:
        return [
            titlecase(case_name)
            for case_name in self.html.xpath(
                f'//*[@id="{self.year}"]/../following-sibling::section/table//tr/td[3]/text()'
            )
        ]

    def _get_case_dates(self) -> List[datetime.date]:
        self.year = self.html.xpath('//li[@class="menuitems"]/a/text()')[-1]
        return [
            convert_date_string(
                date_text.replace("AUGUSET", "August").replace(
                    "OCOTBER", "October"
                ),
                fuzzy=True,
            )
            for date_text in self.html.xpath(
                f'//*[@id="{self.year}"]/../following-sibling::section/table//tr/td[1]/text()'
            )
        ]

    def _get_precedential_statuses(self) -> List[str]:
        return len(self.download_urls) * ["Published"]



"""
Court Contact: https://www.supremecourt.gov/contact/contact_webmaster.aspx
"""


from datetime import date

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    justices = {
        "A": "Samuel Alito",
        "AB": "Amy Coney Barrett",
        "AS": "Antonin Scalia",
        "B": "Stephen Breyer",
        "BK": "Brett Kavanaugh",
        "D": "Decree",
        "DS": "David Souter",
        "EK": "Elana Kagan",
        "G": "Ruth Bader Ginsburg",
        "JS": "John Paul Stephens",
        "K": "Anthony Kennedy",
        "KJ": "Ketanji Brown Jackson",
        "NG": "Neil Gorsuch",
        "PC": "Per Curiam",
        "R": "John G. Roberts",
        "SS": "Sonia Sotomayor",
        "T": "Clarence Thomas",
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.yy = self._get_current_term()
        self.status = "Published"
        self.url_base = "https://www.supremecourt.gov/opinions"
        self.precedential = "Published"
        self.court = "slipopinion"
        self.url = f"{self.url_base}/{self.court}/{self.yy}"

    @staticmethod
    def _get_current_term():
        """The URLs for SCOTUS correspond to the term, not the calendar.

        The terms kick off on the first Monday of October, so we use October 1st
        as our cut off date.
        """
        today = date.today()
        term_cutoff = date(today.year, 10, 1)
        if today < term_cutoff:
            # Haven't hit the cutoff, return previous year.
            return int(today.strftime("%y")) - 1  # y3k bug!
        else:
            return today.strftime("%y")

    def _process_html(self):
        for row in self.html.xpath("//tr"):
            cells = row.xpath(".//td")
            if len(cells) != 6:
                continue
            a, date, docket, link, justice, citation = row.xpath(".//td")
            if not link.text_content():
                continue
            self.cases.append(
                {
                    "citation": citation.text_content(),
                    "date": date.text_content(),
                    "url": link.xpath(".//a/@href")[0],
                    "name": link.text_content(),
                    "docket": docket.text_content(),
                    "judge": self.justices[justice.text_content()],
                }
            )



"""
Scraper for Indiana Court of Appeals
CourtID: indctapp
Court Short Name: Ind. Ct. App.
Auth: Jon Andersen <janderse@gmail.com>
Reviewer: mlr
History:
    2014-09-03: Created by Jon Andersen
"""
from juriscraper.opinions.united_states.state import ind


class Site(ind.Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.court_name = "Court of Appeals"



import time
from datetime import date

from lxml import html

from juriscraper.OpinionSite import OpinionSite


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = (
            "http://www.armfor.uscourts.gov/newcaaf/opinions/2004Term.htm"
        )
        self.court_id = self.__module__

    def _get_case_names(self):
        return [t for t in self.html.xpath("//table//tr/td[1]/font/text()")]

    def _get_download_urls(self):
        download_urls = []
        for t in self.html.xpath(
            "//table//table//tr[position() > 1]/td[2]/div/font/a[2]/@href"
        ):
            download_urls.append(t)
        return download_urls

    def _get_case_dates(self):
        dates = []
        for e in self.html.xpath(
            "//table/tr[3]/td[2]/blockquote/table/tbody/tr[position() > 1]/td[3]"
        ):
            s = html.tostring(e, method="text", encoding="unicode")
            s = s.replace("June", "Jun").replace("July", "Jul")
            s = " ".join(s.split())
            dates.append(
                date.fromtimestamp(time.mktime(time.strptime(s, "%b %d, %Y")))
            )
        return dates

    def _get_docket_numbers(self):
        docket_numbers = []
        for t in self.html.xpath(
            "//table/tr[3]/td[2]/blockquote/table/tbody/tr/td[2]/div/font/a[1]/text()"
        ):
            docket_numbers.append(t)
        return docket_numbers

    def _get_precedential_statuses(self):
        return ["Published"] * len(self.case_names)

    def _get_citations(self):
        neutral_citations = []
        for e in self.html.xpath(
            "//table/tr[3]/td[2]/blockquote/table/tbody/tr[position() > 1]/td[4]"
        ):
            s = html.tostring(e, method="text", encoding="unicode")
            neutral_citations.append(s)
        return neutral_citations



# History:
#   2015-10-20: Created by Andrei Chelaru

from datetime import datetime

from dateutil import parser
from lxml import html

from juriscraper.AbstractSite import logger
from juriscraper.OpinionSite import OpinionSite


class Site(OpinionSite):
    def __init__(self):
        super().__init__()
        self.year = 0
        self.court_id = self.__module__
        self.url = ""
        self.url_base = "http://www.illinoiscourts.gov/Opinions/SupremeCourt/{year}/default.asp"
        self.back_scrape_iterable = list(range(1996, 2015))
        self.download_url_path = None
        self.case_name_path = None
        self.case_dates_path = None
        self.case_dates_path_day = None
        self.case_dates_path_month = None
        self.precedential_statuses_path = None
        self.docket_numbers_path = None
        self.neutral_citations_path = None

    def _get_download_urls(self):
        return list(self.html.xpath(self.download_url_path))

    def _get_case_names(self):
        case_names = [
            " ".join(i.strip() for i in e.xpath(".//text()")).strip()
            for e in self.html.xpath(self.case_name_path)
        ]
        return case_names

    def _get_case_dates(self):
        if self.year <= 2006:
            case_dates = []
            for e in self.html.xpath(self.case_dates_path):
                month = "".join(e.xpath(self.case_dates_path_month)).strip()
                day = "".join(e.xpath(self.case_dates_path_day)).strip()
                month = month.replace("Decemberr", "December")
                try:
                    case_dates.append(
                        datetime.strptime(
                            f"{self.year} {month} {day}",
                            "%Y %B %d",
                        )
                    )
                except ValueError:
                    raise
        else:
            case_dates = [
                parser.parse(
                    "".join(i.strip() for i in d.xpath(".//text()")),
                    fuzzy=True,
                )
                for d in self.html.xpath(self.case_dates_path)
            ]

        return case_dates

    def _get_precedential_statuses(self):
        statuses = []
        for e in self.html.xpath(self.precedential_statuses_path):
            s = html.tostring(e, method="text", encoding="unicode")
            if "NRel" in s:
                statuses.append("Unpublished")
            else:
                statuses.append("Published")
        return statuses

    def _get_docket_numbers(self):
        docket_numbers = []
        for e in self.html.xpath(self.docket_numbers_path):
            nums = " ".join(i.strip() for i in e.xpath(".//text()"))
            nums = nums.replace("Official Reports", "")
            docket_numbers.append(nums)
        return docket_numbers

    def _get_citations(self):
        neutral_citations = None
        if self.neutral_citations_path:
            neutral_citations = []
            for e in self.html.xpath(self.neutral_citations_path):
                neutral_citations.append(
                    " ".join(i.strip() for i in e.xpath(".//text()")).strip()
                )

        return neutral_citations

    def _download_backwards(self, d):
        self.__set_paths(d)

        self.url = self.url_base.format(year=self.year)
        logger.info(f"Scraping year {self.year}")
        self.html = self._download()
        if self.html is not None:
            # Setting status is important because it prevents the download
            # function from being run a second time by the parse method.
            self.status = 200

    def __set_paths(self, year):
        self.year = year
        if year < 2006:
            base_path = "//tr/td[4][a]"
            self.download_url_path = f"{base_path}/a[1]/@href"
            self.case_name_path = f"{base_path}/preceding-sibling::td[1]"
            self.case_dates_path = base_path
            self.case_dates_path_day = "./preceding-sibling::td[3]//text()"
            self.case_dates_path_month = ".//ancestor::tr[1]/preceding-sibling::tr[count(./td) = 1][td//strong][1]/td//text()"
            self.precedential_statuses_path = (
                f"{base_path}/preceding-sibling::td[2]"
            )
            self.docket_numbers_path = f"{base_path}/preceding-sibling::td[2]"
            self.neutral_citations_path = None
        elif 2006 <= year <= 2010:
            if year == 2010:
                base_path = "//tr/td[3][div/a]"
                self.download_url_path = f"{base_path}/div[1]/a[1]/@href"
                self.case_name_path = f"{base_path}/div[1]/a[1]"
            else:
                base_path = "//tr/td[3][a]"
                self.download_url_path = f"{base_path}/a[1]/@href"
                self.case_name_path = f"{base_path}/a[1]"
            if year == 2006:
                self.case_dates_path = base_path
                self.case_dates_path_day = "./preceding-sibling::td[2]//text()"
                self.case_dates_path_month = ".//ancestor::tr[1]/preceding-sibling::tr[count(./td) = 1][1]/td//text()"
            else:
                self.case_dates_path = f"{base_path}/preceding-sibling::td[2]"
            self.precedential_statuses_path = (
                f"{base_path}/preceding-sibling::td[1]"
            )
            self.docket_numbers_path = f"{base_path}/preceding-sibling::td[1]"
            self.neutral_citations_path = None
        elif 2010 < year < 2015:
            base_path = "//tr/td[4][div/a/text()]"
            self.download_url_path = f"{base_path}/div[1]/a[1]/@href"
            self.case_name_path = f"{base_path}/div[1]/a[1]"
            self.case_dates_path = f"{base_path}/preceding-sibling::td[3]"
            self.precedential_statuses_path = (
                f"{base_path}/preceding-sibling::td[2]"
            )
            self.docket_numbers_path = f"{base_path}/preceding-sibling::td[2]"
            self.neutral_citations_path = "{}/preceding-sibling::td[1]".format(
                base_path
            )



from juriscraper.opinions.united_states.federal_appellate import ca4


class Site(ca4.Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.parameters = {
            "offset": 0,
            "sortBy": "2",
            "query": "collection:(USCOURTS) AND courtname:(United States District Court Eastern District of Louisiana)",
            "historical": True,
        }



# Scraper for Iowa Supreme Court
# CourtID: iowa
# Court Short Name: iowa

from juriscraper.AbstractSite import logger
from juriscraper.lib.string_utils import convert_date_string
from juriscraper.OpinionSite import OpinionSite


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.cases = []
        self.archive = False
        self.back_scrape_iterable = [
            "placeholder"
        ]  # this array can't be empty
        self.url = "https://www.iowacourts.gov/iowa-courts/supreme-court/supreme-court-opinions/"

    def _download(self, request_dict={}):
        html = super()._download(request_dict)
        self.extract_cases(html)
        if self.test_mode_enabled() or self.archive:
            return html

        # Walk over pagination "Next" page(s), if present
        proceed = True
        while proceed:
            next_page_url = self.extract_next_page_url(html)
            if next_page_url:
                logger.info(f"Scraping next page: {next_page_url}")
                html = self._get_html_tree_by_url(next_page_url)
                self.extract_cases(html)
            else:
                proceed = False

    def _get_case_names(self):
        return [case["name"] for case in self.cases]

    def _get_download_urls(self):
        return [case["url"] for case in self.cases]

    def _get_case_dates(self):
        return [convert_date_string(case["date"]) for case in self.cases]

    def _get_precedential_statuses(self):
        return ["Published"] * len(self.case_dates)

    def _get_docket_numbers(self):
        return [case["docket"] for case in self.cases]

    def _download_backwards(self, _):
        """Walk over all "Archive" links on Archive page,
        extract cases dictionaries, and add to self.cases
        """
        self.archive = True
        self.url = f"{self.url}opinions-archive/"
        landing_page_html = self._download()
        path = '//div[@class="main-content-wrapper"]//a[contains(./text(), "Opinions Archive")]/@href'
        for archive_page_url in landing_page_html.xpath(path):
            logger.info(f"Back scraping archive page: {archive_page_url}")
            archive_page_html = self._get_html_tree_by_url(archive_page_url)
            self.extract_archive_cases(archive_page_html)

    def extract_cases(self, html):
        """Extract case dictionaries from "Recent" html page
        and add them to self.cases
        """
        case_substring = "Case No."
        case_elements = html.xpath(f'//h3[contains(., "{case_substring}")]')
        for case_element in case_elements:
            text = case_element.text_content()
            parts = text.split(":")
            docket = parts[0].replace(case_substring, "").strip()
            name = parts[1].strip()
            date_text = case_element.xpath("./following::p[1]")[
                0
            ].text_content()
            date_string = date_text.replace("Filed", "")
            url = case_element.xpath("./following::p[2]//a/@href")[0]
            self.cases.append(
                {
                    "name": name,
                    "docket": docket,
                    "date": date_string,
                    "url": url,
                }
            )

    def extract_archive_cases(self, html):
        """Extract case dictionaries from "Archive" html page
        and add them to self.cases
        """
        path_date = '//div[@class="cms_category_icon_title_row"]'
        for date_header in html.xpath(path_date):
            text = date_header.text_content()
            date_string = text.replace("- DELETE", "")
            path_cases = './following::div[@class="cms_items"][1]/div[@class="cms_item"]'
            for case_container in date_header.xpath(path_cases):
                docket_element = case_container.xpath(
                    './div[@class="cms_item_icon_title_row"]'
                )[0]
                self.cases.append(
                    {
                        "date": date_string,
                        "url": docket_element.xpath(".//a/@href")[0],
                        "docket": docket_element.text_content().strip(),
                        "name": case_container.xpath(
                            './div[@class="cms_item_description"]'
                        )[0]
                        .text_content()
                        .strip(),
                    }
                )

    def extract_next_page_url(self, html):
        """Return the href url from "Next" pagination element
        if it exists, otherwise return False.
        """
        path = '//div[contains(./@class, "pagination-next-page")]//a/@href'
        elements = html.xpath(path)
        return elements[0] if elements else False



from juriscraper.opinions.united_states.federal_appellate import scotus_slip

"""
Court Contact: https://www.supremecourt.gov/contact/contact_webmaster.aspx
"""


from datetime import date

from juriscraper.AbstractSite import logger
from juriscraper.lib.exceptions import InsanityException
from juriscraper.lib.string_utils import convert_date_string
from juriscraper.OpinionSite import OpinionSite


class Site(OpinionSite):
    required_headers = ["Date", "Docket", "Name", "J."]
    expected_headers = required_headers + ["Revised", "R-", "Pt."]
    justices = {
        "A": "Samuel Alito",
        "AB": "Amy Coney Barrett",
        "AS": "Antonin Scalia",
        "B": "Stephen Breyer",
        "BK": "Brett Kavanaugh",
        "D": "Decree",
        "DS": "David Souter",
        "EK": "Elana Kagan",
        "G": "Ruth Bader Ginsburg",
        "JS": "John Paul Stephens",
        "K": "Anthony Kennedy",
        "KJ": "Ketanji Brown Jackson",
        "NG": "Neil Gorsuch",
        "PC": "Per Curiam",
        "R": "John G. Roberts",
        "SS": "Sonia Sotomayor",
        "T": "Clarence Thomas",
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.yy = self._get_current_term()
        self.back_scrape_iterable = list(range(5, int(self.yy) + 1))
        self.url_base = "https://www.supremecourt.gov/opinions"
        self.path_table = "//table[@class='table table-bordered']"
        self.path_row = f"{self.path_table}/tr[position() > 1]"
        self.precedential = "In-chambers"
        self.court = "in-chambers"
        self.headers = False
        self.url = False
        self.headers = []
        self.cases = []

    @staticmethod
    def _get_current_term():
        """The URLs for SCOTUS correspond to the term, not the calendar.

        The terms kick off on the first Monday of October, so we use October 1st
        as our cut off date.
        """
        today = date.today()
        term_cutoff = date(today.year, 10, 1)
        if today < term_cutoff:
            # Haven't hit the cutoff, return previous year.
            return int(today.strftime("%y")) - 1  # y3k bug!
        else:
            return today.strftime("%y")

    def _download(self, request_dict={}):
        if not self.test_mode_enabled():
            self.set_url()
        html = super()._download(request_dict)
        self.extract_cases_from_html(html)
        return html

    def set_url(self):
        self.url = f"{self.url_base}/{self.court}.aspx"

    def set_table_headers(self, html):
        # Do nothing if table is missing
        if html.xpath(self.path_table):
            path = f"{self.path_table}//th"
            self.headers = [
                cell.text_content().strip() for cell in html.xpath(path)
            ]
            # Ensure that expected/required headers are present
            if not set(self.required_headers).issubset(self.headers):
                raise InsanityException("Required table column missing")

    def extract_cases_from_html(self, html):
        self.set_table_headers(html)
        for row in html.xpath(self.path_row):
            case = self.extract_case_data_from_row(row)
            if case:
                # Below will raise key error is new judge key encountered (new SC judge appointed)
                case["judge"] = self.justices[case["J."]] if case["J."] else ""
                self.cases.append(case)
                for revision_data in case["revisions"]:
                    revision = case.copy()
                    revision["Date"] = revision_data["date_string"]
                    revision["Name_Url"] = revision_data["href"]
                    self.cases.append(revision)

    def extract_case_data_from_row(self, row):
        cell_index = 0
        case = {"revisions": []}
        # Process each cell in row
        for cell in row.xpath("./td"):
            text = cell.text_content().strip()
            # Skip rows with blank first cell
            if cell_index == 0 and not text:
                break
            label = self.headers[cell_index]
            if label in ["R-", "Pt."]:
                # Ignore some columns that we don't need
                pass
            elif label == "Revised":
                # It is possible for an opinion to have
                # multiple revisions, so we need to iterate
                # over the links the the cell
                for anchor in cell.xpath("a"):
                    case["revisions"].append(
                        {
                            "href": anchor.xpath("@href")[0],
                            "date_string": anchor.text_content(),
                        }
                    )
            else:
                # Handle normal data cells
                case[label] = text
                href = cell.xpath("./a/@href")
                if href:
                    case[f"{label}_Url"] = href[0]
            cell_index += 1
        return case

    def _get_case_names(self):
        return [case["Name"] for case in self.cases]

    def _get_download_urls(self):
        return [case["Name_Url"] for case in self.cases]

    def _get_case_dates(self):
        return [convert_date_string(case["Date"]) for case in self.cases]

    def _get_docket_numbers(self):
        return [case["Docket"] for case in self.cases]

    def _get_judges(self):
        return [case["judge"] for case in self.cases]

    def _get_precedential_statuses(self):
        return [self.precedential] * len(self.cases)

    def _download_backwards(self, d):
        self.yy = str(d if d >= 10 else f"0{d}")
        logger.info(f"Running backscraper for year: 20{self.yy}")
        self.html = self._download()



"""
Scraper for Maryland Court of Special Appeals
CourtID: mdctspecapp
Court Short Name: MD
Author: Andrei Chelaru
Date created: 06/27/2014
"""


from datetime import date

from juriscraper.opinions.united_states.state import md


class Site(md.Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = "https://www.mdcourts.gov/cgi-bin/indexlist.pl?court=cosa&year={current_year}&order=bydate&submit=Submit".format(
            current_year=date.today().year
        )
        self.court_id = self.__module__



from juriscraper.lib.log_tools import make_default_logger


class JuriscraperException(Exception):
    """
    Base class for Juriscraper custom exceptions
    """

    logger = make_default_logger()


class SkipRowError(JuriscraperException):
    """
    Raise when a row or record has to be skipped
    For example, when collecting opinions and finding an order
    """

    def __init__(self, message: str):
        self.logger.debug(message)


class SlownessException(Exception):
    """Raised when things are too slow."""

    def __init__(self, message):
        Exception.__init__(self, message)


class ParsingException(Exception):
    """Raised when parsing fails."""

    def __init__(self, message):
        Exception.__init__(self, message)


class InsanityException(Exception):
    """Raised when data validation fails."""

    def __init__(self, message):
        Exception.__init__(self, message)


class PacerLoginException(Exception):
    """Raised when the system cannot authenticate with PACER"""

    def __init__(self, message):
        Exception.__init__(self, message)



"""
Scraper for Illinois Appellate Court
CourtID: illappct
Author: Krist Jin
History:
  2013-08-18: Created.
  2014-07-17: Updated by mlr to remedy InsanityException.
  2021-11-02: Updated by satsuki-chan: Updated to new page design.
  2022-01-21: Updated by satsuki-chan: Added validation when citation is missing.
"""

from juriscraper.opinions.united_states.state import ill


class Site(ill.Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = (
            "https://www.illinoiscourts.gov/top-level-opinions?type=appellate"
        )
        self.docket_re = r"(?P<citation>\d{4}\s+IL App\s+(\((?P<district>\d+)\w{1,2}\)\s+)?(?P<docket>\d+\w{1,2})-?U?[BCD]?)"

    def _get_docket(self, match):
        raw_docket = match.group("docket")
        district = match.group("district")
        return f"{district}-{raw_docket[0:2]}-{raw_docket[2:]}"



"""Parse the iquery.pl ("Query" menu) result.

This is pretty limited metadata about the case, although it
presents some more information for BK cases.
"""
import pprint
import re
import sys

from ..lib.log_tools import make_default_logger
from ..lib.string_utils import clean_string, force_unicode, harmonize
from ..lib.utils import clean_court_object
from .docket_report import BaseDocketReport
from .reports import BaseReport

logger = make_default_logger()


class CaseQuery(BaseDocketReport, BaseReport):
    """Parse the iquery.pl ("Query" menu) result.

    This is pretty limited metadata about the case, although it
    presents some more information for BK cases.
    """

    PATH = "cgi-bin/iquery.pl"

    CACHE_ATTRS = ["metadata"]

    ERROR_STRINGS = BaseReport.ERROR_STRINGS + [
        r"Case not found\.",
        r"Only applications for admission to practice",
        r"Only requests for a Certificate of Good Standing",
        r"DO NOT POST IN THIS CASE",
        r"Control F",
    ]

    def __init__(self, court_id, pacer_session=None):
        BaseDocketReport.__init__(self, court_id)
        BaseReport.__init__(self, court_id, pacer_session)

        # Initialize the empty cache properties
        self._clear_caches()
        self._metadata = None

    def parse(self):
        self._clear_caches()
        super().parse()

    @property
    def metadata(self):
        if self.is_valid is False:
            return {}

        if self._metadata is not None:
            return self._metadata

        # The data we're after look like this (re-spacing):
        #
        # <div id="cmecfMainContent">
        #   <input type="hidden" id="cmecfMainContentScroll" value="0">
        #   <CENTER>
        #       <B><FONT SIZE=+1>1:11-cv-10230-MLW</FONT></B>
        #        Arkansas Teacher Retirement System v. State Street
        #         Corporation et al
        #     <BR>
        #       Mark L. Wolf, presiding
        #     <BR>
        #       <B>Date filed:</B> 02/10/2011
        #     <BR>
        #       <B>Date terminated:</B> 06/23/2014
        #     <BR>
        #       <B>Date of last filing:</B> 06/06/2018
        #     <BR>
        #   </CENTER>
        #
        # There's a bit more structured data in bankruptcy cases;
        # but note the extra <B></B> pair in the first line:
        #
        # <div id="cmecfMainContent">
        #   <input type="hidden" id="cmecfMainContentScroll" value="0">
        #   <CENTER>
        #       <B><FONT SIZE=+1>18-11572</FONT></B>
        #       <B></B>Nancy Jean Stevens
        #     <BR>
        #       <B>Case type:</B> bk
        #       <B>Chapter:</B> 7
        #       <B>Asset:</B> No
        #       <B>Vol: </B> v
        #       <b>Judge:</b> Frank J. Bailey
        #     <BR>
        #       <B>Date filed:</B> 04/30/2018
        #       <B>Date of last filing:</B> 06/04/2018
        #     <BR>
        #   </CENTER>
        #
        # There is some BK variation, both in terms of presence of
        # docket numbers suffixes, and how the Judge: field name
        # varies depending on the judge's chief judge status:
        #
        # <div id="cmecfMainContent">
        #   <input type="hidden" id="cmecfMainContentScroll" value="0">
        #   <CENTER>
        #       <B><FONT SIZE=+1>18-11573</FONT></B>
        #       <B></B>Sara A Taylor
        #     <BR>
        #       <B>Case type:</B> bk
        #       <B>Chapter:</B> 7
        #       <B>Asset:</B> No
        #       <B>Vol: </B> v
        #       <b>Chief Judge:</b> Melvin S. Hoffman
        #     <BR>
        #       <B>Date filed:</B> 04/30/2018
        #       <B>Date of last filing:</B> 06/06/2018
        #     <BR>
        #   </CENTER>

        # Rather than take the approach used by DocketParser of
        # searching throughout the document for a docket number by
        # regular expression, instead we take the <center> tag that follows
        # the relevant cmecfMainContent <div>, and go line-by-line,
        # delimited by <br>s. This approach is more structured:
        center_path = './/div[@id="cmecfMainContent"]//center'
        try:
            center = self.tree.xpath(center_path)[0]
        except IndexError:
            # Can happen for sealed cases? We have at least one test case here.
            self._metadata = {}
            return {}

        rows = self.redelimit_p(center, self.BR_REGEX)

        # The first row demands special handling:
        #   <B><FONT SIZE=+1>18-11572</FONT></B><B></B>Nancy Jean Stevens
        # We take the docket number from the <font> tag (the innermost tag),
        # although we could but have chosen the first <b> tag.
        docket_number = self._parse_docket_number_strs(
            [rows[0].find(".//font").text_content()]
        )
        # And case caption following the final <b></b> pair.
        case_name_raw = force_unicode(rows[0].find(".//b[last()]").tail or "")
        if not case_name_raw:
            # In some courts, such as INSB, the case name is within a font tag.
            # <b><font size="+1">13-07422-RLM-7A</font></b><font size="+1">
            #  Cedric Morris and Kateen Nicklondra Morris
            # </font>
            try:
                case_name_raw = force_unicode(
                    rows[0].xpath(
                        ".//b[last()]/following-sibling::font[1]/text()"
                    )[0]
                )
            except IndexError:
                case_name_raw = ""

        # Our job as a parser is to return the data, not to filter, clean,
        # amend, or "harmonize" it. However downstream often expects harmonized
        # data, so go with both.
        case_name = clean_string(harmonize(case_name_raw))

        # Iterate through the remaining rows, recognizing that the second row
        # may be special (district court), in which case it lacks the
        # <b> tag used for key/value pairs and it gives the judge's name.
        # Such as:
        #   Mark L. Wolf, presiding
        #   <B>Date filed:</B> 02/10/2011
        #   <B>Date filed:</B> 04/30/2018<B>Date of last filing:</B> 06/06/2018
        data = {}
        field_names = {
            "date_of_last_filing": "date_last_filing",
            "judge": "assigned_to_str",
            "chief_judge": "assigned_to_str",
            "plan_confirmed": "date_plan_confirmed",
            "debtor_discharged": "date_debtor_dismissed",
            "joint_debtor_discharged": "date_joint_debtor_dismissed",
        }
        for i in range(1, len(rows) - 1):
            bolds = rows[i].findall(".//b")
            if not bolds:
                line = force_unicode(rows[i].text_content())
                if not line:
                    continue

                if re.search(r"(panel [2-3]$|Probation)", line):
                    # Skip panelists after the first, skip probation
                    # assignments
                    continue

                if i <= 3:
                    # Second & third rows without bold => judge name!
                    presiding_re = re.compile(", (presiding|panel 1)$")
                    referral_re = re.compile(", referral$")
                    related_re = re.search("Related case: (.*)", line)
                    if presiding_re.search(line):
                        data["assigned_to_str"] = presiding_re.sub("", line)
                    elif referral_re.search(line):
                        data["referred_to_str"] = referral_re.sub("", line)
                    elif related_re:
                        data["related_cases_str"] = related_re.group(1)
                    else:
                        raise AssertionError(
                            f"Unable to match judge row: {line}"
                        )
                else:
                    raise AssertionError(f"Line with no boldface: '{line}'")
            for bold in bolds:
                data.update(
                    self._get_label_value_pair(bold, True, field_names)
                )

        data.update(
            {
                "court_id": self.court_id,
                "docket_number": docket_number,
                "case_name": case_name,
                "case_name_raw": case_name_raw,
            }
        )

        data = clean_court_object(data)

        self._metadata = data
        return data

    def query(self, pacer_case_id):
        """Use a district court's PACER query function with a known case ID

        At the top of every district PACER court, there's a button that says,
        "Query". When you click that button, it either does a query for the
        criteria you specify, returning a list of results or it returns the
        specific case you were looking for, including metadata and a list of
        reports you can run on that case.

        While there is utility in both of these result types, this method only
        supports the use case where you know the pacer_case_id in advance, and
        are expecting only one result in return. This method does *not* support
        parsing the search results that the Query button can return. That use
        case is supposed by the CaseQueryAdvancedBankruptcy and
        CaseQueryAdvancedDistrict objects.

        :param pacer_case_id: A pacer_case_id for a case to lookup.
        :return None: Instead, sets self.response attribute and runs
        self.parse()
        """

        """
        This is an example of a query that could be sent, and which we're
        attempting to replicate:
            curl 'https://ecf.ord.uscourts.gov/cgi-bin/iquery.pl?1-L_1_0-1'
                -H 'Content-Type: multipart/form-data; boundary=----WebKitFormBoundary7Sz5UdWuoi5TbtDj'
                -H 'Referer: https://external'
                -H 'Cookie: CASE_NUM=6:16-cv-01710(128548); PacerSession=xxx; PacerPref=receipt=Y'
                --data-binary $'------WebKitFormBoundary7Sz5UdWuoi5TbtDj\r\nContent-Disposition: form-data;
                    name="UserType"\r\n\r\n\r\n------WebKitFormBoundary7Sz5UdWuoi5TbtDj\r\nContent-Disposition: form-data;
                    name="all_case_ids"\r\n\r\n128548\r\n------WebKitFormBoundary7Sz5UdWuoi5TbtDj\r\nContent-Disposition: form-data;
                    name="case_num"\r\n\r\n6:16-cv-1710\r\n------WebKitFormBoundary7Sz5UdWuoi5TbtDj\r\nContent-Disposition: form-data;
                    name="Qry_filed_from"\r\n\r\n\r\n------WebKitFormBoundary7Sz5UdWuoi5TbtDj\r\nContent-Disposition: form-data;
                    name="Qry_filed_to"\r\n\r\n\r\n------WebKitFormBoundary7Sz5UdWuoi5TbtDj\r\nContent-Disposition: form-data;
                    name="lastentry_from"\r\n\r\n\r\n------WebKitFormBoundary7Sz5UdWuoi5TbtDj\r\nContent-Disposition: form-data;
                    name="lastentry_to"\r\n\r\n\r\n------WebKitFormBoundary7Sz5UdWuoi5TbtDj\r\nContent-Disposition: form-data;
                    name="last_name"\r\n\r\n\r\n------WebKitFormBoundary7Sz5UdWuoi5TbtDj\r\nContent-Disposition: form-data;
                    name="first_name"\r\n\r\n\r\n------WebKitFormBoundary7Sz5UdWuoi5TbtDj\r\nContent-Disposition: form-data;
                    name="middle_name"\r\n\r\n\r\n------WebKitFormBoundary7Sz5UdWuoi5TbtDj\r\nContent-Disposition: form-data;
                    name="person_type"\r\n\r\n\r\n------WebKitFormBoundary7Sz5UdWuoi5TbtDj--\r\n'

        The important notes, from testing:
         - The CASE_NUM cookie isn't needed.
         - The case_num parameter needs a value; any value will do.
         - As usual, the ?1-L_1_0-1 business in the URL is needed
        """
        assert (
            self.session is not None
        ), "session attribute of CaseQuery report cannot be None."
        assert bool(
            pacer_case_id
        ), f"pacer_case_id must be truthy, not '{pacer_case_id}'"
        params = {
            "UserType": "",
            "all_case_ids": pacer_case_id,
            "case_num": "foo",  # We just need *some* value here.
            "Qry_filed_from": "",
            "Qry_filed_to": "",
            "lastentry_from": "",
            "lastentry_to": "",
            "last_name": "",
            "first_name": "",
            "middle_name": "",
            "person_type": "",
        }
        logger.info(
            "Running case query for case ID '%s' in court '%s'",
            pacer_case_id,
            self.court_id,
        )
        self.response = self.session.post(f"{self.url}?1-L_1_0-1", data=params)
        self.parse()

    @property
    def data(self):
        """Get all the data back from this endpoint.

        Don't attempt to return parties or docket_entries like the superclass
        does.
        """
        if self.is_valid is False:
            return {}

        data = self.metadata.copy()
        return data


def _main():
    if len(sys.argv) != 2:
        print("Usage: python -m juriscraper.pacer.case_query filepath")
        print("Please provide a path to an HTML file to parse.")
        sys.exit(1)

    # Court ID is only needed for querying. Actual
    # parsed value appears in output
    report = CaseQuery("mad")
    filepath = sys.argv[1]
    print(f"Parsing HTML file at {filepath}")
    with open(filepath) as f:
        text = f.read().decode("utf-8")
    report._parse_text(text)
    pprint.pprint(report.data, indent=2)


if __name__ == "__main__":
    _main()



"""Scraper for Missouri
CourtID: mo
Court Short Name: MO
Author: Ben Cassedy
Date created: 04/27/2014
"""
from datetime import date

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.court = "Supreme"
        self.url = self.build_url()
        self.status = "Published"

    def build_url(self):
        year = date.today().year
        return f"https://www.courts.mo.gov/page.jsp?id=12086&dist=Opinions%20{self.court}&date=all&year={year}#all"

    def _process_html(self):
        for row in self.html.xpath("//div[@class='margin-bottom-15']"):
            date = row.xpath(".//input")[0].value
            for opinion in row.xpath(".//div[@class='list-group-item-text']"):
                links = opinion.xpath("a")
                if len(links) != 2:
                    continue
                url = opinion.xpath("a")[1].get("href")
                all_text = opinion.xpath(".//text()")
                case_metadata = [t.strip() for t in all_text if t.strip()]
                docket, _, name, _, author, _, vote = case_metadata
                self.cases.append(
                    {
                        "name": name,
                        "docket": docket[:-1],
                        "url": url,
                        "date": date,
                        "dispostion": vote.split(".")[0].strip(),
                        "judge": author,
                        "judges": vote.split(".", 1)[1].strip(),
                    }
                )



"""Scraper for the United States Department of Justice Attorney General
CourtID: ag
Court Short Name: United States Attorney General
"""
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "https://www.justice.gov/olc/opinions?items_per_page=40"
        self.status = "Published"

    def _process_html(self):
        for row in self.html.xpath(".//article"):
            name = row.xpath(".//h2")[0].text_content().strip()
            url = row.xpath(".//a/@href")[0]
            date = row.xpath(".//time")[0].text_content()
            if not name:
                continue
            summary = row.xpath(".//p")[0].text_content()
            self.cases.append(
                {
                    "date": date,
                    "name": name,
                    "url": url,
                    "summary": summary,
                    "docket": "",  # Docket numbers don't appear to exist.
                }
            )



"""Scraper for Second Circuit
CourtID: ca2
Author: MLR
Reviewer: MLR
Contact:
  Webmaster@ca2.uscourts.gov ('Calendar Team' handles updates, and is responsive)
  ecfhelpdesk@ca2.uscourts.gov
  Shane_Clouden@ca2.uscourts.gov
History:
  2016-09-09: Created by MLR
  2023-11-21: Fixed by flooie
  2023-12-11: Fixed by quevon24
"""

from lxml import html

from juriscraper.OralArgumentSiteLinear import OralArgumentSiteLinear


class Site(OralArgumentSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.base_url = "http://www.ca2.uscourts.gov"
        self.url = f"{self.base_url}/decisions"
        self.method = "POST"
        self.base_xpath = '//tr[contains(.//a/@href, "mp3")]'
        self.parameters = {
            "IW_SORT": "-DATE",
            "IW_BATCHSIZE": "50",
            "IW_FILTER_DATE_BEFORE": "",
            "IW_FILTER_DATE_AFTER": "NaNNaNNaN",
            "IW_FIELD_TEXT": "*",
            "IW_DATABASE": "Oral Args",
            "opinion": "*",
        }
        self.backscraper = False
        self.back_scrape_iterable = ["placeholder"]

    def _process_html(self):
        for row in self.html.xpath("//table[@border='1']"):
            link = row.xpath(".//a")[0].get("href")
            if ".mp3" not in link:
                continue  # skip bad data
            docket = row.xpath(".//a/nobr/text()")[0]
            name, date = row.xpath(".//td/text()")
            url = self._build_url(link)
            if docket == "GMT20230614-175136_Recording":
                continue  # skip bad data
            self.cases.append(
                {"docket": docket, "url": url, "name": name, "date": date}
            )

    def _download(self, request_dict={}):
        if self.test_mode_enabled():
            self.html = super()._download()
            return self.html
        if self.backscraper:
            return super()._download()
        r = self.request["session"].post(self.url, params=self.parameters)
        return html.fromstring(r.content)

    def _build_url(self, link):
        if self.backscraper:
            url = link
        else:
            url = f"{self.base_url}{link}"
        return url

    def _download_backwards(self, d):
        self.backscraper = True
        self.method = "GET"
        for i in range(1, 33):
            self.url = f"{self.base_url}/decisions/isysquery/19342efa-bb49-4684-a054-875324b58eb9/{(i*10)-9}-{i*10}/list/"
            self.html = self._download()
            self._process_html()



from juriscraper.AbstractSite import AbstractSite


class OpinionSite(AbstractSite):
    """Contains generic methods for scraping data. Should be extended by all
    scrapers.

    Should not contain lists that can't be sorted by the _date_sort function.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Order of attributes is important as it affects the order of parsing.
        # Some methods rely on others having already been run.
        self._opt_attrs = [
            "adversary_numbers",
            "causes",
            "dispositions",
            "divisions",
            "docket_attachment_numbers",
            "docket_document_numbers",
            "docket_numbers",
            "judges",
            "lower_courts",
            "lower_court_judges",
            "lower_court_numbers",
            "nature_of_suit",
            "citations",
            "parallel_citations",
            "summaries",
            "case_name_shorts",
            "child_courts",
        ]
        self._req_attrs = [
            "case_dates",
            "case_names",
            "download_urls",
            "precedential_statuses",
            "blocked_statuses",
            "date_filed_is_approximate",
        ]
        # For date sorting to work, case_dates must be first in _all_attrs.
        self._all_attrs = self._req_attrs + self._opt_attrs

        # Set all metadata to None
        for attr in self._all_attrs:
            self.__setattr__(attr, None)

    def _get_adversary_numbers(self):
        # Common in bankruptcy cases where there are adversary proceedings.
        return None

    def _get_download_urls(self):
        raise NotImplementedError(
            "`_get_download_urls()` must be implemented."
        )

    def _get_case_dates(self):
        raise NotImplementedError("`_get_case_dates()` must be implemented.")

    def _get_case_names(self):
        raise NotImplementedError("`_get_case_names()` must be implemented.")

    def _get_causes(self):
        return None

    def _get_dispositions(self):
        return None

    def _get_divisions(self):
        return None

    def _get_docket_attachment_numbers(self):
        return None

    def _get_docket_document_numbers(self):
        return None

    def _get_docket_numbers(self):
        return None

    def _get_judges(self):
        return None

    def _get_nature_of_suit(self):
        return None

    def _get_citations(self):
        return None

    def _get_parallel_citations(self):
        return None

    def _get_lower_courts(self):
        return None

    def _get_lower_court_judges(self):
        return None

    def _get_lower_court_numbers(self):
        return None

    def _get_precedential_statuses(self):
        raise NotImplementedError(
            "`_get_precedential_statuses()` must be implemented."
        )

    def _get_summaries(self):
        return None

    def _get_child_courts(self):
        return None

    def extract_from_text(self, scraped_text):
        """Pass scraped text into function and return data as a dictionary

        :param opinion_text: Text of scraped content
        :return: dictionary of information
        """
        return {}

    def _get_date_filed_is_approximate(self):
        return [False] * len(self.case_names)



"""Scraper for Dept of Justice Board of Immigration Appeals
CourtID: bia
Court Short Name: Dept of Justice Board of Immigration Appeals
Author: William Palin
Reviewer:
Type:
History:
    2021-12-18: Created by William E. Palin
"""
import re
from datetime import datetime
from typing import Any, Dict

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "https://www.justice.gov/eoir/ag-bia-decisions"
        self.volume = 0
        self.urls = None
        self.status = "Published"

    def _process_html(self):
        if not self.test_mode_enabled():
            if not self.urls:
                urls = self.html.xpath(
                    ".//table[1]/tbody/tr/td/a[contains(., 'Volume')]"
                )

                def get_text(elem):
                    return elem.text_content()

                self.urls = sorted(urls, key=get_text, reverse=True)
            self.url = self.urls[self.volume].get("href")
            self.html = super()._download()
        table = self.html.xpath(".//table")[0]
        for row in table.xpath(".//strong"):
            name = row.text_content().strip().strip(",")
            row_text = row.xpath("..")[0].text_content()
            if "BIA" not in row_text:
                continue
            if not name:
                continue
            citation, year = row_text.split(name)[1].split("(")
            cells = row.xpath("..")[0].xpath(
                "following-sibling::td[position() <= 2]"
            )
            if not cells:
                continue
            url = cells[0].xpath(".//a/@href")
            docket = cells[0].xpath(".//a")[0].text_content()
            if not url:
                continue
            self.cases.append(
                {
                    "name": name,
                    "citation": citation,
                    "url": url[0],
                    "docket": docket,
                    "date": f"{year.split()[1]}-07-01",
                    "date_filed_is_approximate": True,
                }
            )

    def extract_from_text(self, scraped_text: str) -> Dict[str, Any]:
        """Can we extract the date filed from the text?

        :param scraped_text: The content of the document downloaded
        :return: Metadata to be added to the case
        """
        date = re.findall(
            r"Decided (by (Acting\s)?Attorney General )?(.*\d{4})",
            scraped_text,
        )[0][-1]
        date_filed = datetime.strptime(date, "%B %d, %Y").strftime("%Y-%m-%d")
        metadata = {
            "OpinionCluster": {
                "date_filed": date_filed,
                "date_filed_is_approximate": False,
            },
        }
        return metadata



"""
Scraper for the Tennessee Court of Criminal Appeals
CourtID: tennctcrimapp
Court Short Name: Tenn. Ct. Crim. App.
"""
from juriscraper.opinions.united_states.state import tenn


class Site(tenn.Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = (
            "https://www.tncourts.gov/courts/court-criminal-appeals/opinions"
        )



# Scraper for Louisiana Supreme Court
# CourtID: la
# Court Short Name: LA
# Contact: Community relations department
#          Robert Gunn
#          504-310-2592
#          rgunn@lasc.org

from datetime import date

from juriscraper.lib.html_utils import get_html_parsed_text
from juriscraper.lib.string_utils import titlecase
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.year = date.today().year
        self.url = "http://www.lasc.org/CourtActions/%d" % self.year
        self.status = "Published"

    def _download(self, request_dict={}):
        landing_page = OpinionSiteLinear._download(self, request_dict)
        if self.test_mode_enabled():
            return [self._get_subpage_html_by_page(landing_page)]
        path = (
            "//tr["
            "contains(./td[3], 'Opinion') or "
            "contains(./td[3], 'PER CURIAM')"
            "]//td[2]//@href"
        )
        # parse 2 most recent Opinon/PerCuriam sub-pages
        urls = landing_page.xpath(path)[:2]
        return [self._get_subpage_html_by_url(url) for url in urls]

    def _process_html(self):
        path = (
            "//a["
            "contains(., 'v.') or "
            "contains(., 'IN RE') or "
            "contains(., 'IN THE') or "
            "contains(., 'vs.') or "
            "contains(., 'VS.')"
            "]"
        )
        for html in self.html:
            for anchor in html.xpath(path):
                date_string = self._get_date_for_opinions(html)
                text = anchor.text_content()
                parts = text.split(None, 1)
                summary_lines = anchor.getparent().xpath("./text()")
                self.cases.append(
                    {
                        "date": date_string,
                        "docket": parts[0],
                        "judge": self._get_judge_above_anchor(anchor),
                        "name": titlecase(parts[1]),
                        "summary": " ".join(summary_lines).replace(text, ""),
                        "url": f"http://www.lasc.org{anchor.get('href')}",
                    }
                )

    def _get_date_for_opinions(self, html):
        element_date = html.xpath("//span")[0]
        element_date_text = element_date.text_content().strip()
        parts = element_date_text.split("day of")
        day = parts[0].split()[-1]
        month = parts[1].split()[0]
        year = parts[1].split()[1].strip(",")
        return " ".join([month, day, year])

    def _get_judge_above_anchor(self, anchor):
        path = (
            "./preceding::*["
            "starts-with(., 'BY ') or "
            "contains(., 'CURIAM:')"
            "]"
        )
        try:
            text = anchor.xpath(path)[-1].text_content()
        except IndexError:
            return None
        return text.rstrip(":").lstrip("BY").strip()

    def _get_subpage_html_by_url(self, url):
        page = self._get_html_tree_by_url(url)
        return self._get_subpage_html_by_page(page)

    def _get_subpage_html_by_page(self, page):
        path = ".//textarea[@id='PostContent']"
        html = page.xpath(path)[0].text_content()
        return get_html_parsed_text(html)



import codecs
import os
import unittest

from setuptools import find_packages, setup
from setuptools.command.install import install

VERSION = "2.5.82"
AUTHOR = "Free Law Project"
EMAIL = "info@free.law"
HERE = os.path.abspath(os.path.dirname(__file__))


reqs_path = f"{HERE}/requirements.txt"
with open(reqs_path) as reqs_file:
    reqs = reqs_file.read().splitlines()


def read(*parts):
    """
    Build an absolute path from *parts* and and return the contents of the
    resulting file.  Assume UTF-8 encoding.
    """
    with codecs.open(os.path.join(HERE, *parts), "rb", "utf-8") as f:
        return f.read()


class TestNetwork(install):
    """Run network test only"""

    description = "run isolated tests that hit the network"

    def run(self):
        loader = unittest.defaultTestLoader
        runner = unittest.TextTestRunner(verbosity=2)
        tests = loader.discover("./tests/network")
        runner.run(tests)


setup(
    name="juriscraper",
    description="An API to scrape American court websites for metadata.",
    license="BSD",
    url="https://github.com/freelawproject/juriscraper",
    version=VERSION,
    author=AUTHOR,
    author_email=EMAIL,
    maintainer=AUTHOR,
    maintainer_email=EMAIL,
    keywords=["scraping", "legal", "pacer"],
    long_description=read("README.rst"),
    long_description_content_type="text/x-rst",
    packages=find_packages(exclude=["tests*"]),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    install_requires=reqs,
    tests_require=["jsondate3-aware"],
    include_package_data=True,
    test_suite="tests.test_local",
    cmdclass={
        "testnetwork": TestNetwork,
    },
)



# Scraper for Superior Court of the Virgin Islands (published)
# CourtID: visuper_p

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = (
            "https://superior.vicourts.org/court_opinions/published_opinions"
        )
        self.status = "Published"

    def _process_html(self):
        for row in self.html.xpath(".//tbody//tr"):
            (
                case_name,
                date_filed,
                docket_number,
                judge,
                citation,
                _summary,
            ) = row.xpath(".//td//text()")
            url = row.xpath(".//td/a")[0].get("href")
            self.cases.append(
                {
                    "date": date_filed,
                    "docket": docket_number,
                    "name": case_name,
                    "url": url,
                    "judge": judge,
                    "citation": citation,
                }
            )



import json
import re

import requests
from requests.packages.urllib3 import exceptions

from ..lib.exceptions import PacerLoginException
from ..lib.html_utils import get_html_parsed_text, get_xml_parsed_text
from ..lib.log_tools import make_default_logger
from ..pacer.utils import is_pdf, is_text

logger = make_default_logger()

requests.packages.urllib3.disable_warnings(exceptions.InsecureRequestWarning)


def check_if_logged_in_page(content: bytes) -> bool:
    """Is this a valid HTML page from PACER?

    Check if the data in 'content' is from a valid PACER page or valid PACER
    XML document, or if it's from a page telling you to log in or informing you
    that you're not logged in.
    :param content: The data to test, of type bytes. This uses bytes to avoid
    converting data to text using an unknown encoding. (see #564)
    :return boolean: True if logged in, False if not.
    """

    valid_case_number_query = (
        b"<case number=" in content
        or b"<request number=" in content
        or b'id="caseid"' in content
        or b"Cost: " in content
    )
    no_results_case_number_query = re.search(b"<message.*Cannot find", content)
    sealed_case_query = re.search(b"<message.*Case Under Seal", content)
    if any(
        [
            valid_case_number_query,
            no_results_case_number_query,
            sealed_case_query,
        ]
    ):
        not_logged_in = re.search(b"text.*Not logged in", content)
        if not_logged_in:
            # An unauthenticated PossibleCaseNumberApi XML result. Simply
            # continue onwards. The complete result looks like:
            # <request number='1501084'>
            #   <message text='Not logged in.  Please refresh this page.'/>
            # </request>
            # An authenticated PossibleCaseNumberApi XML result.
            return False
        else:
            return True

    # Detect if we are logged in. If so, no need to do so. If not, we login
    # again below.
    found_district_logout_link = b"/cgi-bin/login.pl?logout" in content
    found_appellate_logout_link = b"InvalidUserLogin.jsp" in content

    # A download confirmation page doesn't contain a logout link but we're
    # logged into.
    is_a_download_confirmation_page = b"Download Confirmation" in content
    # When looking for a download confirmation page sometimes an appellate
    # attachment page is returned instead, see:
    # https://ecf.ca8.uscourts.gov/n/beam/servlet/TransportRoom?servlet=ShowDoc&pacer=i&dls_id=00802251695
    appellate_attachment_page = (
        b"Documents are attached to this filing" in content
    )
    # Sometimes the document is completely unavailable and an error message is
    # shown, see:
    # https://ecf.ca11.uscourts.gov/n/beam/servlet/TransportRoom?servlet=ShowDoc/009033568259
    appellate_document_error = (
        b"The requested document cannot be displayed" in content
    )
    if any(
        [
            found_district_logout_link,
            found_appellate_logout_link,
            is_a_download_confirmation_page,
            appellate_attachment_page,
            appellate_document_error,
        ]
    ):
        # A normal HTML page we're logged into.
        return True

    return False


class PacerSession(requests.Session):
    """
    Extension of requests.Session to handle PACER oddities making it easier
    for folks to just POST data to PACER endpoints/apis.

    Also includes utilities for logging into PACER and re-logging in when
    sessions expire.
    """

    LOGIN_URL = "https://pacer.login.uscourts.gov/services/cso-auth"

    def __init__(
        self, cookies=None, username=None, password=None, client_code=None
    ):
        """
        Instantiate a new PACER API Session with some Juriscraper defaults
        :param cookies: an optional RequestsCookieJar object with cookies for the session
        :param username: a PACER account username
        :param password: a PACER account password
        :param client_code: an optional PACER client code for the session
        """
        super().__init__()
        self.headers["User-Agent"] = "Juriscraper"
        self.headers["Referer"] = "https://external"  # For CVE-001-FLP.
        self.verify = False

        if cookies:
            assert not isinstance(cookies, str), (
                "Got str for cookie parameter. Did you mean "
                "to use the `username` and `password` kwargs?"
            )
            self.cookies = cookies

        self.username = username
        self.password = password
        self.client_code = client_code
        self.additional_request_done = False

    def get(self, url, auto_login=True, **kwargs):
        """Overrides request.Session.get with session retry logic.

        :param url: url string to GET
        :param auto_login: Whether the auto-login procedure should happen.
        :return: requests.Response
        """
        kwargs.setdefault("timeout", 300)

        r = super().get(url, **kwargs)

        if b"This user has no access privileges defined." in r.content:
            # This is a strange error that we began seeing in CM/ECF 6.3.1 at
            # ILND. You can currently reproduce it by logging in on the central
            # login page, selecting "Court Links" as your destination, and then
            # loading: https://ecf.ilnd.uscourts.gov/cgi-bin/WrtOpRpt.pl
            # The solution when this error shows up is to simply re-run the get
            # request, so that's what we do here. PACER needs some frustrating
            # and inelegant hacks sometimes.
            r = super().get(url, **kwargs)
        if auto_login:
            updated = self._login_again(r)
            if updated:
                # Re-do the request with the new session.
                r = super().get(url, **kwargs)
                # Do an additional check of the content returned.
                self._login_again(r)
        return r

    def post(self, url, data=None, json=None, auto_login=True, **kwargs):
        """
        Overrides requests.Session.post with PACER-specific fun.

        Will automatically convert data dict into proper multi-part form data
        and pass to the files parameter instead.

        Will set a timeout of 300 if not provided.

        All other uses or parameters will pass through untouched
        :param url: url string to post to
        :param data: post data
        :param json: json object to post
        :param auto_login: Whether the auto-login procedure should happen.
        :param kwargs: assorted keyword arguments
        :return: requests.Response
        """
        kwargs.setdefault("timeout", 300)

        if data:
            pacer_data = self._prepare_multipart_form_data(data)
            kwargs.update({"files": pacer_data})
        else:
            kwargs.update({"data": data, "json": json})

        r = super().post(url, **kwargs)
        if auto_login:
            updated = self._login_again(r)
            if updated:
                # Re-do the request with the new session.
                return super().post(url, **kwargs)
        return r

    def head(self, url, **kwargs):
        """
        Overrides request.Session.head with a default timeout parameter.

        :param url: url string upon which to do a HEAD request
        :param kwargs: assorted keyword arguments
        :return: requests.Response
        """
        kwargs.setdefault("timeout", 300)
        return super().head(url, **kwargs)

    @staticmethod
    def _prepare_multipart_form_data(data):
        """
        Transforms a data dictionary into the multi-part form data that PACER
        expects as the POST body
        :param data: dict of data to transform
        :return: dict with values wrapped into tuples like:(None, <value>)
        """
        output = dict()
        for key in data:
            output[key] = (None, data[key])
        return output

    @staticmethod
    def _get_view_state(r):
        """Get the viewState parameter of the form

        This is an annoying thing we have to do. The login flow has three
        requests that you make and each requires the view state from the one
        prior. Thus, we capture that viewState each time and submit it during
        each of the next submissions.

        The HTML takes the form of:

        <input type="hidden" name="javax.faces.ViewState"
               id="j_id1:javax.faces.ViewState:0"
               value="some-long-value-here">

        :param r: A request.Response object
        :return The value of the "value" attribute of the ViewState input
        element.
        """
        tree = get_html_parsed_text(r.content)
        xpath = (
            '//form[@id="loginForm"]//input['
            '    @name="javax.faces.ViewState"'
            "]/@value"
        )
        return tree.xpath(xpath)[0]

    @staticmethod
    def _get_xml_view_state(r):
        """Same idea as above, but sometimes PACER returns XML so we parse
        that instead of the HTML.

        Here's a sample of the XML:

        <partial-response id="j_id1">
          <changes>
            <update id="regmsg:bpmConfirm">
              <![CDATA[<button id="regmsg:bpmConfirm" name="regmsg:bpmConfirm" class="ui-button ui-widget ui-state-default ui-corner-all ui-button-text-only" onclick="PrimeFaces.ab({s:&quot;regmsg:bpmConfirm&quot;,pa:[{name:&quot;dialogName&quot;,value:&quot;redactionDlg&quot;}]});return false;" style="margin-right: 20px;" type="submit"><span class="ui-button-text ui-c">Continue</span></button><script id="regmsg:bpmConfirm_s" type="text/javascript">PrimeFaces.cw("CommandButton","widget_regmsg_bpmConfirm",{id:"regmsg:bpmConfirm"});</script>]]>
            </update>
            <update id="loginForm:pclLoginMessages">
              <![CDATA[<div id="loginForm:pclLoginMessages" class="ui-messages ui-widget" style="font-size: 0.85em;" aria-live="polite"></div>]]>
            </update>
            <update id="j_id1:javax.faces.ViewState:0"><![CDATA]]>
            </update>
          </changes>
        </partial-response>
        """
        tree = get_xml_parsed_text(r.content)
        xpath = "//update[@id='j_id1:javax.faces.ViewState:0']/text()"
        return tree.xpath(xpath)[0]

    def login(self, url=None):
        """Attempt to log into the PACER site.
        The first step is to get an authentication token using a PACER
        username and password.
        To get the authentication token, it's necessary to send a POST request:
        curl --location --request POST 'https://pacer.login.uscourts.gov/services/cso-auth' \
            --header 'Accept: application/json' \
            --header 'User-Agent: Juriscraper' \
            --header 'Content-Type: application/json' \
            --data-raw '{
                "loginId": "USERNAME",
                "password": "PASSWORD"
            }'

        All documentation for PACER Authentication API User Guide can be found here:
        https://pacer.uscourts.gov/help/pacer/pacer-authentication-api-user-guide
        """
        logger.info("Attempting PACER API login")
        # Clear any remaining cookies. This is important because sometimes we
        # want to login before an old session has entirely died.
        self.cookies.clear()
        if url is None:
            url = self.LOGIN_URL
        # By default, it's assumed that the user is a filer. Redaction flag is set to 1
        data = {
            "loginId": self.username,
            "password": self.password,
            "redactFlag": "1",
        }
        # If optional client code information is included, include in login request
        if self.client_code:
            data["clientCode"] = self.client_code

        login_post_r = super().post(
            url,
            headers={
                "User-Agent": "Juriscraper",
                "Content-type": "application/json",
                "Accept": "application/json",
            },
            timeout=60,
            data=json.dumps(data),
        )

        if login_post_r.status_code != requests.codes.ok:
            message = f"Unable connect to PACER site: '{login_post_r.status_code}: {login_post_r.reason}'"
            logger.warning(message)
            raise PacerLoginException(message)

        # Continue with login when response code is "200: OK"
        response_json = login_post_r.json()

        # 'loginResult': '0', user successfully logged; '1', user not logged
        if (
            response_json.get("loginResult") == None
            or response_json.get("loginResult") == "1"
        ):
            message = f"Invalid username/password: {response_json.get('errorDescription')}"
            raise PacerLoginException(message)
        # User logged, but with pending actions for their account
        if response_json.get("loginResult") == "0" and response_json.get(
            "errorDescription"
        ):
            logger.info(response_json.get("errorDescription"))

        if not response_json.get("nextGenCSO"):
            raise PacerLoginException(
                "Did not get NextGenCSO cookie when attempting PACER login."
            )
        # Set up cookie with 'nextGenCSO' token (128-byte string of characters)
        session_cookies = requests.cookies.RequestsCookieJar()
        session_cookies.set(
            "NextGenCSO",
            response_json.get("nextGenCSO"),
            domain=".uscourts.gov",
            path="/",
        )
        # Support "CurrentGen" servers as well. This can be remoevd if they're
        # ever all upgraded to NextGen.
        session_cookies.set(
            "PacerSession",
            response_json.get("nextGenCSO"),
            domain=".uscourts.gov",
            path="/",
        )
        # If optional client code information is included,
        # 'PacerClientCode' cookie should be set
        if self.client_code:
            session_cookies.set(
                "PacerClientCode",
                self.client_code,
                domain=".uscourts.gov",
                path="/",
            )
        self.cookies = session_cookies
        logger.info("New PACER session established.")

    def _do_additional_request(self, r: requests.Response) -> bool:
        """Check if we should do an additional request to PACER, sometimes
        PACER returns the login page even though cookies are still valid.
        Do an additional GET request if we haven't done it previously.
        See https://github.com/freelawproject/courtlistener/issues/2160.

        :param r: The requests Response object.
        :return: True if an additional request should be done, otherwise False.
        """
        if r.request.method == "GET" and self.additional_request_done is False:
            self.additional_request_done = True
            return True
        return False

    def _login_again(self, r):
        """Log into PACER if the session has credentials and the session has
        expired.

        :param r: A response object to inspect for login errors.
        :returns: A boolean indicating whether a new session needed to be
        created.
        :raises: PacerLoginException, if unable to create a new session.
        """
        if is_pdf(r):
            return False

        if is_text(r):
            return False

        logged_in = check_if_logged_in_page(r.content)
        if logged_in:
            return False

        if self.username and self.password:
            logger.info(
                "Invalid/expired PACER session. Establishing new session."
            )
            self.login()
            return True
        else:
            if self._do_additional_request(r):
                return True
            raise PacerLoginException(
                "Invalid/expired PACER session and do not have credentials "
                "for re-login."
            )



"""Scraper for the Supreme Court of Michigan
CourtID: mich
Court Short Name: Mich.
Contact: sitefeedback@courts.mi.gov
History:
    - 2014-09-21: Updated by Jon Andersen to handle some fields being missing
    - 2014-08-05: Updated to have a dynamic URL, an oversight during check in.
    - 2022-01-28: Updated for new web site, @satsuki-chan.
"""
import re
from typing import List
from urllib.parse import urlencode

from juriscraper.DeferringList import DeferringList
from juriscraper.lib.string_utils import titlecase
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.filters = (
            ("releaseDate", "Past Year"),
            ("page", "1"),
            ("sortOrder", "Newest"),
            ("searchQuery", ""),
            ("resultType", "opinions"),
            ("pageSize", "100"),
        )
        self.court = "Supreme Court"
        params = self.filters + (
            ("resource", "supreme-court-opinion"),
            ("aAppellateCourt", self.court),
        )
        self.url = f"https://www.courts.michigan.gov/api/CaseSearch/SearchCaseOpinions?{urlencode(params)}"
        # Currently blocked from Michigan so we drop the user agent for now
        self.request["headers"] = {"User-Agent": ""}
        self.title_re = r"(MSC|COA) (?P<docket>\d+)\s+(?P<name>.+)\s+Opinion"
        self.status = "Published"
        self.cases = []

    def _process_html(self) -> None:
        """Process the html and extract out the opinions

        :return: None
        """
        for item in self.html["searchItems"]:
            match = re.search(self.title_re, item["title"])
            if match:
                docket = match.group("docket")
                name = self.cleanup_case_name(match.group("name"))
            else:
                # In some cases the case name is missing. Grab the docket from the title in a different manner
                docket = item["title"].split("_")[0]
                name = ""
            self.cases.append(
                {
                    "date": item["displayDate"],
                    "docket": docket,
                    "name": name,
                    "url": f"https://www.courts.michigan.gov{item['documentUrl'].strip()}",
                    "lower_court": self.get_lower_courts(item["courts"]),
                    "title": item["title"],
                }
            )

    def _get_case_names(self) -> List[str]:
        """Get case names *if* missing

        In some cases the case name is missing. This method uses a deferred list to the get the case name.

        :return: List of case names
        """

        def fetcher(case):
            if case["name"] != "":
                # Return the name we extracted without using fetcher
                return case["name"]
            elif self.test_mode_enabled():
                # if we're in test mode, return a dummy name
                return "Test Name"
            else:
                # Else, query the API and return the name of the case
                self.url = f"https://www.courts.michigan.gov/api/CaseSearch/SearchCaseSearchContent/?searchQuery={case['title']}"
                self.html = self._download()
                case["name"] = self.html["opinionResults"]["searchItems"][0][
                    "title"
                ].title()
            return case["name"]

        return DeferringList(seed=self.cases, fetcher=fetcher)

    def cleanup_case_name(self, name_raw: str) -> str:
        """Clean up case name in Michigan

        :param name_raw: Raw title string
        :return: Cleaned name
        """
        name = titlecase(name_raw)
        if "People of Mi " in name:
            name = name.replace("People of Mi ", "People of Michigan ")
        return name

    def get_lower_courts(self, courts: List[str]) -> str:
        """Get the lower courts

        :param courts: Court names as an array
        :return: The lower courts combined into a string
        """
        if courts:
            lower_courts = ", ".join(courts).lstrip(", ")
        else:
            lower_courts = ""
        return titlecase(lower_courts)



from datetime import date, timedelta

from juriscraper.lib.html_utils import get_table_column_text
from juriscraper.lib.string_utils import convert_date_string
from juriscraper.OpinionSite import OpinionSite


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        begin_date = date.strftime(date.today() - timedelta(15), "%m/%d/%Y")
        end_date = date.strftime(date.today(), "%m/%d/%Y")
        self.url = (
            "http://wicourts.gov/supreme/scopin.jsp?"
            "begin_date=%s&end_date=%s&SortBy=date"
        ) % (begin_date, end_date)
        self.court_id = self.__module__
        self.table_id = "scopinion"
        self.path_base = f"//table[@id='{self.table_id}']"

    def _get_case_names(self):
        return get_table_column_text(self.html, 3, False, self.table_id)

    def _get_download_urls(self):
        path = f"{self.path_base}//td[4]//a[contains(., 'PDF')]/@href"
        return self.html.xpath(path)

    def _get_case_dates(self):
        path = f"{self.path_base}//td[1]"
        cells = self.html.xpath(path)
        return [convert_date_string(cell.text_content()) for cell in cells]

    def _get_docket_numbers(self):
        return get_table_column_text(self.html, 2, False, self.table_id)

    def _get_precedential_statuses(self):
        return ["Published"] * len(self.case_names)



"""
History:
 - 2014-08-05: Adapted scraper to have year-based URLs.
 - 2023-11-18: Fixed and updated
"""

from juriscraper.opinions.united_states.state import orctapp


class Site(orctapp.Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = (
            "https://www.courts.oregon.gov/publications/sc/Pages/default.aspx"
        )
        self.status = "Published"
        self.court_code = "p17027coll3"



"""Scraper for Supreme Court of Virgin Islands
CourtID: virginislands
Court Short Name: Virgin Islands
Author: William Edward Palin
History:
  2023-01-21: Created by William Palin
"""
import urllib.parse
from datetime import date, datetime, timedelta

from juriscraper.lib.string_utils import convert_date_string
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = (
            f"https://supreme.vicourts.org/court_opinions/published_opinions"
        )
        self.status = "Published"
        today = date.today()
        self.previous_date = today - timedelta(days=60)

    def _process_html(self):
        if self.test_mode_enabled():
            self.previous_date = datetime(2023, 9, 21).date()

        for s in self.html.xpath(".//tr/td/.."):
            cells = s.xpath(".//td")
            if not cells[0].text_content():
                continue

            judges = cells[3].text_content()
            name = cells[0].text_content()
            date = cells[1].text_content().replace("/", "-")
            date_object = convert_date_string(date)
            docket = cells[2].text_content()
            citation = cells[4].text_content()
            if date_object < self.previous_date:
                continue
            u = s.xpath(".//td/a/@href")[0]
            url = urllib.parse.quote(u, safe="/:")
            self.cases.append(
                {
                    "name": name,
                    "date": date,
                    "docket": docket,
                    "citation": citation,
                    "judges": judges,
                    "url": url,
                }
            )



import os

from selenium import webdriver
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import Select, WebDriverWait

from juriscraper.lib.cookie_utils import normalize_cookies
from juriscraper.lib.html_utils import (
    clean_html,
    fix_links_but_keep_anchors,
    get_html_parsed_text,
)


class WebDriven:
    def __init__(self, *args, **kwargs):
        self.cookies = {}
        self.url = False
        self.uses_selenium = True
        self.wait = False  # type: WebDriverWait
        self.webdriver = False  # type: webdriver

    def __del__(self):
        self.close_webdriver_session()

    def action_chain_click(self, element: WebElement):
        """Try this when you are getting an error like:
        Element is not clickable at point (x, y) because
        another element obscures it
        """
        ActionChains(self.webdriver).move_to_element(element).click().perform()

    def close_webdriver_session(self):
        if self.webdriver:
            # sometimes quiting below throws a cryptic and otherwise ignorable OS error
            try:
                self.webdriver.quit()
            except:
                pass

    def dump_html_web_element(self, element: WebElement):
        """Use this for debugging purposes"""
        print(element.get_attribute("innerHTML"))

    def find_element(self, by: str, path: str) -> WebElement:
        if self.webdriver:
            return self.webdriver.find_element(by, path)
        raise Exception("webdriver not initiated")

    def find_element_by_class_name(self, path: str) -> WebElement:
        return self.find_element(By.CLASS_NAME, path)

    def find_element_by_id(self, path: str) -> WebElement:
        return self.find_element(By.ID, path)

    def find_element_by_xpath(self, path: str) -> WebElement:
        return self.find_element(By.XPATH, path)

    def get_page(self) -> WebElement:
        text = clean_html(self.webdriver.page_source)
        html = get_html_parsed_text(text)
        html.rewrite_links(fix_links_but_keep_anchors, base_href=self.url)
        return html

    def initiate_webdriven_session(self):
        if not self.url:
            raise Exception("self.url not set")

        options = webdriver.FirefoxOptions()
        if os.environ.get("SELENIUM_VISIBLE", False):
            options.headless = False
        else:
            options.headless = True
        options.accept_insecure_certs = True
        webdriver_conn = os.environ.get("WEBDRIVER_CONN", "local")
        if webdriver_conn == "local":
            # See README instruction for installing geckodriver
            self.webdriver = webdriver.Firefox(options=options)
        else:
            capabilities = options.to_capabilities()
            self.webdriver = webdriver.Remote(
                webdriver_conn,
                desired_capabilities=capabilities,
                keep_alive=True,
            )

        self.webdriver.implicitly_wait(30)
        self.webdriver.set_window_size(5000, 10000)
        self.wait = WebDriverWait(self.webdriver, 20)
        self.webdriver.get(self.url)
        self.cookies = normalize_cookies(self.webdriver.get_cookies())

    def scroll_to_element_then_click(self, element: WebElement):
        script = "arguments[0].click();"
        webdriver.Firefox.execute_script(script, element)

    def select_form_option(self, form_id: str) -> Select:
        element = self.find_element_by_id(form_id)
        return Select(element)

    def select_form_option_value(self, form_id: str, form_value: str):
        self.select_form_option(form_id).select_by_value(form_value)

    def select_form_option_text(self, form_id: str, form_value_text: str):
        self.select_form_option(form_id).select_by_visible_text(
            form_value_text
        )

    def wait_for_id_then_click(self, id: str):
        self.wait.until(EC.element_to_be_clickable((By.ID, id))).click()

    def wait_for_path_then_click(self, xpath: str):
        self.wait.until(EC.element_to_be_clickable((By.XPATH, xpath))).click()

    def wait_for_id(self, id_attr):
        self.wait.until(EC.presence_of_element_located((By.ID, id_attr)))

    def take_screenshot(self, name: str = None):
        """Use this method to snap screenshots during debugging"""
        name = name if name else f"screenshot.{self.__module__}.png"
        self.webdriver.save_screenshot(name)



#!/usr/bin/python
import argparse
import glob
import json
import pprint

from lxml import etree

__author__ = "mlissner"


def xpath(tree, query):
    return tree.xpath(
        query,
        namespaces={
            "m": "http://www.loc.gov/mods/v3",
            "xlink": "http://www.w3.org/1999/xlink",
        },
    )


def get_court_id(fdsys_court):
    """Map the fdsys_court string to a CourtListener string"""
    return fdsys_court


def clean_up_rough_data(data):
    """Clean up the data so it's good.

    - [ ] Strip extra spaces from all nodes.
    - [ ] Convert newlines into spaces.
    - [ ] Ensure good encodings?
    - [ ] Review other items from Juriscraper.

    """
    return data


def get_unique_nodes(node_list):
    """Lots of duplication in the XML. Nuke duplicates."""
    good_nodes = []
    nodes_as_txt = []
    for node in node_list:
        txt = etree.tostring(node).strip()
        if txt in nodes_as_txt:
            # Duplicate. Move on.
            pass
        else:
            nodes_as_txt.append(txt)
            good_nodes.append(node)
    return good_nodes


def get_parties(tree):
    """Extract the parties from the XML into a nice object."""
    party_nodes = xpath(tree, "//m:party")
    party_nodes = get_unique_nodes(party_nodes)

    parties = []
    for party_node in party_nodes:
        parties.append(
            {
                "name_first": xpath(party_node, "@m:firstName"),
                "name_last": xpath(party_node, "@m:lastName"),
                "name_middle": xpath(party_node, "@m:middleName"),
                "name_suffix": xpath(party_node, "@m:generation"),
                "role": xpath(party_node, "@m:role"),
            }
        )
    return parties


def get_documents(tree):
    """Get the documents from the XML into a nice object."""
    document_nodes = xpath(tree, "//m:mods/m:relatedItem")
    documents = []
    for document_node in document_nodes:
        documents.append(
            {
                "download_url": xpath(
                    document_node, "m:relatedItem/@xlink:href"
                ),
                "description": xpath(document_node, "//m:subTitle"),
                "date_filed": xpath(document_node, "XXX"),
            }
        )


def print_xpath_results():
    for f in glob.glob("../examples/*.xml"):
        tree = etree.parse(f)
        rough_data = {
            "download_url": xpath(
                tree, "(//m:identifier[@type='uri'])[1]/text()"
            ),
            "fdsys_id": xpath(tree, "(//m:accessId/text())[1]"),
            "court_id": get_court_id(xpath(tree, "(//m:courtCode/text())[1]")),
            "docket_number": xpath(tree, "(//m:caseNumber/text())[1]"),
            "court_location": xpath(tree, "(//m:caseOffice/text())[1]"),
            "parties": get_parties(tree),
            "case_name": xpath(tree, "(//m:titleInfo/m:title/text())[1]"),
            "documents": get_documents(tree),
        }
        data = clean_up_rough_data(rough_data)
        pprint.pprint(data)


def main():
    parser = argparse.ArgumentParser(
        description="Get all the metadata from a mods file."
    )
    args = parser.parse_args()

    print_xpath_results()


def get_fdsys_court_names():
    response = glob.glob("../examples/2015_USCOURTS_sitemap.xml")
    tree = etree.parse(response[0])
    # print(etree.tostring(tree, pretty_print=True))
    data = dict()

    for url in tree.xpath(
        "//m:loc/text()",
        namespaces={
            "m": "http://www.sitemaps.org/schemas/sitemap/0.9",
            "xlink": "http://www.w3.org/1999/xlink",
        },
    ):
        pre = url.split("-")[1]
        # if pre not in data and url:
        data[pre] = url

    with open("./fdsys_court_names.json", "w") as f:
        json.dump(data, f)


if __name__ == "__main__":
    main()
    # get_fdsys_court_names()



# Scraper for Georgia Appeals Court
# CourtID: gactapp
# Court Short Name: gactapp
# Author: Andrei Chelaru
# Reviewer: mlr
# Date created: 25 July 2014


from datetime import date, timedelta

from juriscraper.lib.string_utils import titlecase
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        today = date.today().strftime("%Y-%m-%d")
        last_week = (date.today() - timedelta(days=7)).strftime("%Y-%m-%d")
        self.status = "Published"
        self.url = f"https://www.gaappeals.us/wp-content/themes/benjamin/docket/docketdate/results_all.php?OPstartDate={last_week}&OPendDate={today}&submit=Start+Opinions+Search"

    def _process_html(self):
        for row in self.html.xpath("//tr")[::-1][:-1]:
            docket, name, date, disposition, _, url = row.xpath(".//td")
            self.cases.append(
                {
                    "docket": docket.text_content(),
                    "name": titlecase(name.text_content()),
                    "date": date.text_content(),
                    "disposition": disposition.text_content().title(),
                    "url": url.xpath(".//a")[0].get("href"),
                }
            )



import os

from requests import HTTPError


def build_module_list(court_id):
    """Takes a string and builds up a list of modules to import.

    This is a simple recursive function that iteratively looks for __all__
    attributes in packages. If it finds one, it inspects each of the items in
    it to see if each of those has an __all__ attribute. If they lack it, the
    item is added to a list of modules.

    Returns either a list of modules or in the case of errors, an empty list.

    *N.B.*: For the most part, this re-creates the `walk` function from the
    pkgutil library. At some point, it's probably worth replacing this with
    that. Live and learn.

    https://docs.python.org/2/library/pkgutil.html#pkgutil.walk_packages
    """
    module_strings = []

    def find_all_attr_or_punt(court_id):
        """Checks that we have an __all__ attribute. If so, recurses. If not,
        adds the item to our list
        """
        try:
            # Test that we have an __all__ attribute (proving that it's a
            # package) something like: opinions.united_states.federal
            all_attr = __import__(court_id, globals(), locals(), ["*"]).__all__

            # Build the modules back up to full imports
            all_attr = [f"{court_id}.{item}" for item in all_attr]
            for module in all_attr:
                # If we've made it this far, we have an __all__ attribute, so
                # we should see if the items within that attribute do too. And
                # so forth, recursively...
                find_all_attr_or_punt(module)
        except AttributeError:
            # Lacks the __all__ attribute. Probably of the form:
            # juriscraper.opinions.united_states.federal_appellate.ca1,
            # therefore, we add it to our list!
            module_strings.append(court_id)
        # except ImportError as e:
        #    # Something has gone wrong with the import
        #    print("Import error: %s" % e)
        #    return []

    find_all_attr_or_punt(court_id)

    return module_strings


def get_module_by_name(name):
    db_root = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "..", "opinions")
    )
    for dirName, subdirList, fileList in os.walk(db_root):
        for fname in fileList:
            if f"{name}.py" == fname:
                package, module = (
                    dirName.split("/juriscraper/", 1)[1].replace("/", "."),
                    fname[:-3],
                )
                juriscraper_module = __import__(
                    f"juriscraper.{package}.{module}",
                    globals(),
                    locals(),
                    [module],
                )
                return juriscraper_module.Site()


def site_yielder(iterable, mod):
    for i in iterable:
        site = mod.Site()
        try:
            site._download_backwards(i)
            yield site
        except HTTPError as e:
            continue



# Scraper for Kansas Supreme Court (published)
# CourtID: kan_p

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "https://www.kscourts.org/Cases-Decisions/Decisions"
        self.request["verify"] = False
        self.status = "Published"
        self.court = "Supreme Court"

    def _update_parameters(self):
        """"""
        view_state = self.html.xpath("//input[@id='__VIEWSTATE']")[0].get(
            "value"
        )
        VIEWSTATEGENERATOR = self.html.xpath(
            "//input[@id='__VIEWSTATEGENERATOR']"
        )[0].get("value")
        CMSCsrfToken = self.html.xpath("//input[@id='__CMSCsrfToken']")[0].get(
            "value"
        )
        data = {
            "__CMSCsrfToken": CMSCsrfToken,
            "__EVENTTARGET": "",
            "__EVENTARGUMENT": "",
            "lng": "en-US",
            "__VIEWSTATEGENERATOR": VIEWSTATEGENERATOR,
            "__SCROLLPOSITIONX": "0",
            "__SCROLLPOSITIONY": "239",
            "p$lt$ctl01$SmartSearchBox3$txtWord_exWatermark_ClientState": "",
            "p$lt$ctl01$SmartSearchBox3$txtWord": "",
            "p$lt$zonePagePlaceholder$pageplaceholder$p$lt$ctl02$OpinionFilter1$filterControl$txtSearch": "",
            "p$lt$zonePagePlaceholder$pageplaceholder$p$lt$ctl02$OpinionFilter1$filterControl$drpPublished": self.status,
            "p$lt$zonePagePlaceholder$pageplaceholder$p$lt$ctl02$OpinionFilter1$filterControl$drpCourt": self.court,
            "p$lt$zonePagePlaceholder$pageplaceholder$p$lt$ctl02$OpinionFilter1$filterControl$drpSortBy": "Sort By",
            "p$lt$zonePagePlaceholder$pageplaceholder$p$lt$ctl02$OpinionFilter1$filterControl$btnFilter": "Apply Filters",
            "p$lt$zonePagePlaceholder$pageplaceholder$p$lt$ctl03$AccordionLayout1$acc_AccordionExtender_ClientState": "-1",
            "p$lt$zonePagePlaceholder$pageplaceholder$p$lt$ctl07$AccordionLayout$acc_AccordionExtender_ClientState": "-1",
            "p$lt$ctl09$SmartSearchBox1$txtWord_exWatermark_ClientState": "",
            "p$lt$ctl09$SmartSearchBox1$txtWord": "",
            "__VIEWSTATE": view_state,
        }
        self.parameters = data
        self.method = "POST"
        self.request["verify"] = False

    def _process_html(self):
        self.method = "POST"
        if not self.test_mode_enabled():
            if (
                self.url
                == "https://www.kscourts.org/Cases-Decisions/Decisions"
            ):
                self._update_parameters()
                self.html = super()._download()
        for row in self.html.xpath(".//tr"):
            date_filed, docket_number, case_name, court, status = row.xpath(
                ".//td/a/text()"
            )
            url = row.xpath(".//td/a")[0].get("href")
            self.cases.append(
                {
                    "status": status,
                    "date": date_filed,
                    "docket": docket_number,
                    "name": case_name,
                    "url": url,
                }
            )



#  Scraper for Georgia Supreme Court
# CourtID: ga
# Court Short Name: ga

import re
from datetime import date, datetime, timedelta

from juriscraper.AbstractSite import logger
from juriscraper.lib.string_utils import titlecase
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        today = date.today()
        self.url = self._get_url(today.year)
        self.status = "Published"
        self.back_scrape_iterable = range(2016, 2022)

    def _get_url(self, year: int) -> str:
        """Generate the GA URL for a given year.

        :param year: Year to scrape
        :return: URL for the given year
        """
        return f"https://www.gasupreme.us/opinions/{year}-opinions/"

    def _process_html(self) -> None:
        for link in self.html.xpath("//li/a[contains(@href, '.pdf')]"):
            url = link.get("href")
            # Expected title content:
            # - "S20A1505, S20A1506. PENDER v. THE STATE"
            # - "S21A0306. BELL v. RAFFENSPERGER"
            title = link.text_content()
            if not title:
                continue
            dockets = re.findall(r"S?\d{2}\w\d{4}", title)
            if not dockets:
                # Skip links to weekly summaries or cases without a docket
                continue
            docket = ", ".join(dockets)
            name = title.split(dockets[-1])[-1].strip("., ")
            # Expected summary content:
            # - "October 19, 2021—SUMMARIES for NOTEWORTHY OPINIONS"
            # - "July 7, 2021"
            # - "February 15, 2021 – SUMMARIES for NOTEWORTHY OPINIONS"
            summary = link.xpath(
                ".//parent::li/parent::ul/preceding-sibling::p[1]"
            )[0].text_content()
            # Character separator for dates from summary text could be:
            # - dash: "-"
            # - hyphen: "—"
            # - character U+2013: "–"
            date_str = (
                summary.split("–")[0].split("—")[0].split("-")[0].strip()
            )
            date_summary = datetime.strptime(date_str, "%B %d, %Y")
            self.cases.append(
                {
                    "date": date_str,
                    "docket": docket,
                    "name": titlecase(name),
                    "url": url,
                }
            )

    def _download_backwards(self, year) -> None:
        self.url = self._get_url(year)
        logger.info(f"Backscraping for year {year}: {self.url}")
        self.html = self._download()

        # Setting status is important because it prevents the download
        # function from being run a second time by the parse method.
        if self.html is not None:
            self.status = 200
            self._process_html()



"""Scraper for the Texas Attorney General
CourtID: texag
Court Short Name: Texas Attorney General
History:
    2017-02-04: Created by Ardery
    2018-10-13: Retired by Ardery
    2023-01-28: Updated by William E. Palin
"""

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "https://texasattorneygeneral.gov/opinions"

    def _process_html(self):
        cases = self.html.xpath("//div[@class='sidebar-ag-opinion-content']")
        for case in cases:
            docket = case.xpath(".//h4")[0].text_content().strip()
            summary = case.xpath(".//p")[0].text_content().strip()
            url = case.xpath(".//h4/a/@href")[0]
            name = f"Untitled Texas Attorney General Opinion: {docket}"
            date = case.xpath(
                ".//div[@class='sidebar-ag-opinion-casedate']/text()"
            )[0].strip()
            self.cases.append(
                {
                    "name": name,
                    "docket": docket,
                    "date": date,
                    "url": url,
                    "summary": summary,
                    "status": "Published",
                }
            )



"""Backscraper for Supreme Court of Maine 2021 redux
CourtID: me
Court Short Name: Me.
Author: William E. Palin
Date created: March 31, 2021
"""

from juriscraper.opinions.united_states.state import me


class Site(me.Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.back_scrape_iterable = [2020, 2019, 2018, 2017]

    def _download_backwards(self, year):
        self.url = f"https://www.courts.maine.gov/courts/sjc/lawcourt/{year}/index.html"
        self.html = self._download()



"""
Scraper for New Hampshire Supreme Court
CourtID: nh
Court Short Name: NH
Court Contact: webmaster@courts.state.nh.us
Author: Andrei Chelaru
Reviewer: mlr
History:
    - 2014-06-27: Created
    - 2014-10-17: Updated by mlr to fix regex error.
    - 2015-06-04: Updated by bwc so regex catches comma, period, or
    whitespaces as separator. Simplified by mlr to make regexes more semantic.
    - 2016-02-20: Updated by arderyp to handle strange format where multiple
    case names and docket numbers appear in anchor text for a single case pdf
    link. Multiple case names are concatenated, and docket numbers are
    concatenated with ',' delimiter
    - 2021-12-29: Updated for new web site, by flooie and satsuki-chan
"""

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "https://www.courts.nh.gov/content/api/documents?iterate_nodes=true&q=%40field_document_subcategory%7C%3D%7C2256%40field_document_purpose%7C%3D%7C1331&sort=field_date_posted%7Cdesc%7CALLOW_NULLS&filter_mode=exclusive&page=1&size=25"
        self.status = "Published"

    def _download(self):
        if self.test_mode_enabled():
            import json

            return json.load(open(self.url))
        return (
            self.request["session"]
            .get(
                self.url,
                headers=self.request["headers"],
                verify=self.request["verify"],
                timeout=60,
            )
            .json()
        )

    def _process_html(self) -> None:
        for case in self.html["data"]:
            url = case["fields"]["field_document_file"]["0"]["fields"]["uri"][
                0
            ].split("//")[1]
            if "," not in case["title"]:
                continue
            docket, name = case["title"].split(",", 1)
            self.cases.append(
                {
                    "name": name.strip(),
                    "docket": docket.strip(),
                    "date": case["fields"]["field_date_posted"][0],
                    "url": f"https://www.courts.nh.gov/sites/g/files/ehbemt471/files/{url}",
                }
            )
